%{
//  Copyright (c) by RPS Professional Software
//----------------------------------------------------------------------------
//  Project name: SOKO
//  Module file : FSPARSER.Y
//  Compiler    : Borland C++ 4.5
//  Copied from : Original
//  Created by  : Bojic Dragan
//----------------------------------------------------------------------------
//  Description :
//      Yacc input file; FirmStyle & Metrics common analysis parser
//----------------------------------------------------------------------------
//  Changes:
//  $Log__$
//----------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>

#ifdef DEBUG
   #ifdef BRANKO
      #include "debug.h"
   #endif
#endif

/* obrada deklaracija i tabela simbola */
#include "fsbool.h"
#include "fserror.h"
#include "fssymtab.h"

/* leksicki analizator */
#include "fsanzopt.h"
#include "fslex.h"

inline void yyerror( char *str )
{
   if ( strcmp( str, "syntax error" ) != 0 ) FSAnzFatalError( E_PARSER_STACK_OVERFLOW );
}

// specific for pcyacc
#define yychar pcyytoken
#define YYEMPTY -1
#define YYACCEPT return(0)

%}
/* tip elementa semantickog steka */

%union {
	FSSymbol *ptrSym;
	FSType *ptrType;
	FSSymbolList *ptrSymList;
	FSBaseTypeSpecifiers spec;
	FSTypeId typId;
    char *ptrChar;
	char ascii;
}
%type <ptrType> type specifiers opt_specifiers type_or_class type_specifier
%type <ptrSym> ext_decl decl
%type <ptrSym> var_decl funct_decl new_name decl_name enumerator
%type <ptrSym> param_declaration abstract_decl abs_decl arg_decl
%type <ptrSymList> name_list var_list
%type <ptrType> struct_specifier enum_specifier
%type <ptrSym> tag opt_tag struct_decl opt_enum_tag
%type <ptrSymList> struct_def_list enumerator_list
%type <spec> star const_volat_list spec_mod const_volat
/*
%type <ptrChar> expr, non_comma_expr, or_expr, or_list
%type <ptrChar> and_expr, and_list, binary, unary
%type <ptrGraph> statement , stmt_list
*/

/* nonstandard extensions' tokens for specific compilers */
%term <spec> _SPEC_MOD _SPEC_QUAL _SEGMENT
%left <spec> _HW_SPEC_QUAL
%right <spec> _HI_TECH_SPEC_QUAL
%term <spec> _SPEC_STRG_CL
%term <spec> _SBIT _BIT _SFR _SFR16 _SFRP _BASED _IAR_INT _IAR_USING
%term _THREAD _ASM _HI_TECH_ASM _EXCEPT  _FINALLY _TRY _PRIORITY _TASK
%term _USING _AT _COSMIC_6805_AT _C51_NOPAR _C51_166_INT _C166_REGMASK
%term _IAR_UNARY _IAR_ARGS  _IAR_CON _SEGNAME _SELF _EMIT


%term NEWLINE SHARP SHARPSHARP _ASM_START _ASM_END // not used in parser, but preprocessor needs it in fspars_t.h
%term STRING /* String constant */
%term ICON /* Integer or long constant including '\t' etc. */
%term FCON /* Floating-point constant */
%term <spec> TYPE /* int char long float double signed unsigned short */
	   /* void */
%term <spec> CONST_VOLAT
%term <typId> STRUCT /* struct union */
%term ENUM /* enum */
%term <ptrSym> TTYPE /* type created with typedef */
%nonassoc <spec> CLASS /* auto extern register static typedef */
%nonassoc <ptrChar> NAME
%right ELSE
%right error

%term RETURN GOTO
%term IF ELSE
%term SWITCH CASE DEFAULT
%term BREAK CONTINUE
%term WHILE DO FOR
%term LC RC /* {} */
%term SEMI /* ; */
%term ELLIPSIS /* ... */

%left COMMA /* , */
%right <ascii> EQUAL ASSIGNOP /* = *= /= %= += -= <<= >>= &= |= ^= */
%right QUEST COLON /* ? : */
%left OROR /* || */
%left ANDAND /* && */
%left OR /* | */
%left XOR /* ^ */
%left AND /* & */
%left <ascii> EQUOP /* == != */
%left <ascii> RELOP /* <= >= < > */
%left <ascii> SHIFTOP /* >> << */
%left PLUS MINUS /* + - */
%left <ascii> STAR DIVOP /* * / % */
%right <ascii> SIZEOF UNOP INCOP /* sizeof ! ~ ++ -- */
%left <ascii> _BASE_OPERATOR /* :> proveriti prioritet */
%left <ascii> LB RB LP RP STRUCTOP _IAR_CON /* [] () . -> */

%%
	/*******************************************************************
	 *                          The C grammar                          *
	 *******************************************************************/

	/* -------------------High Level Definitions---------------------- */

program
	: ext_def_list {}
;
ext_def_list
	: ext_def_list ext_def
	| /* empty */
		{ /* initializations */
		  typDfAllowed = TRUE;
        FSLexInit();
		}
;
ext_def
	: opt_specifiers error /*error err_end */
      {
         FSAnzError( E_INV_DECL );
         yyerrok;
         if ( pcyytoken == NAME ) delete [] yylval.ptrChar;
         while ( ( pcyytoken = yylex() ) != SEMI && pcyytoken != EOF )
         {
            if ( pcyytoken == NAME ) delete [] yylval.ptrChar;
         }
         if ( pcyytoken == SEMI ) pcyytoken = yylex();
         if ( pcyytoken == EOF ) pcyytoken = 0;
         delete $1;
      }
	| opt_specifiers ext_decl_list SEMI /* no action */
   | opt_specifiers ext_decl_list error
      {
        FSAnzError( E_INV_DECL );
        yyerrok;
      }
	| opt_specifiers { typDfAllowed = TRUE; } SEMI
		{ delete $1;
		}
/*	| opt_specifiers funct_decl error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $2 = FSAppendType( $1, $2);
         fsSymTable->Insert( $2 );
      }
*/	| opt_specifiers funct_decl arg_def_list
		{ $2 = FSAppendType( $1, $2);
          fsSymTable->Insert( $2 );
		  fsSymTable->EnterScope();
		  FSInsertSymListtoTable(*fsSymTable,*( ((FSFunctType *)($2->GetType()))->GetArgs() ));
		}
	  compound_stmt
		{
		  fsSymTable->LeaveScope();
		}
   | _HI_TECH_ASM LP string_const RP SEMI
;

ext_decl_list
	: ext_decl
		{ $1 = FSAppendType( $<ptrType>0, $1, (yychar == COMMA) ? typDfAllowed = FALSE, FS_SPEC_COPY : FS_NO_SPEC_COPY);
         if ( yychar != COMMA ) $<ptrType>0 = 0;
		  fsSymTable->Insert( $1 );
		}
	| ext_decl_list comma_modif ext_decl
		{ $3 = FSAppendType( $<ptrType>0, $3, (yychar == COMMA) ? typDfAllowed = FALSE, FS_SPEC_COPY : FS_NO_SPEC_COPY);
         if ( yychar != COMMA ) $<ptrType>0 = 0;
		  fsSymTable->Insert( $3 );
		}
;
ext_decl
	: var_decl  /* $$ = $1 */
   | var_decl EQUAL error
      {
         FSAnzError( E_INV_DECL );
         yyerrok;
         $1->SetDef( FS_DEFINED );
      }
	| var_decl EQUAL initializer
		{ $1->SetDef(FS_DEFINED);
		}
   | var_decl _AT initializer
		{ $1->SetDef(FS_DEFINED);
		}
   | var_decl _AT error
      {
         FSAnzError( E_INV_DECL );
         yyerrok;
         $1->SetDef( FS_DEFINED );
      }
   | var_decl _COSMIC_6805_AT initializer
		{ $1->SetDef(FS_DEFINED);
		}
   | var_decl _COSMIC_6805_AT error
      {
         FSAnzError( E_INV_DECL );
         yyerrok;
         $1->SetDef( FS_DEFINED );
      }
	| funct_decl %prec error /* $$ = $1; */
   | funct_decl error
      {
         FSAnzError( E_INV_DECL );
         yyerrok;
      } /* for situation when error occurs in K & R args */
;

comma_modif
   : COMMA
   | comma_modif spec_mod
   | comma_modif _HI_TECH_SPEC_QUAL
;

spec_mod
   : _SPEC_MOD
   | _BASED LP unary RP
   | _BASED LP TYPE RP
      { if ( $3 != FS_VOID ) FSAnzError( E_INV_DECL );
      }
;
	/* -------------------------Specifiers----------------------------- */

opt_specifiers
	: specifiers /* $$ = $1 */
	| /* empty */
		{ $$ = new FSBaseType(FS_INT);
		}
;
specifiers
	: type_or_class  /* $$ = $1 */
	| specifiers type_or_class
		{ $$ = FSBaseType::Merge($1,$2);
		}
   | specifiers spec_mod /* $$ = $1 */
;
type
	: type_specifier  /* $$ = $1 */
	| type type_specifier
		{ $$ = FSBaseType::Merge($1,$2);
		}
   | type spec_mod  /* $$ = $1 */
;
type_or_class
	: type_specifier  /* $$ = $1 */
	| CLASS
		{ $$ = new FSBaseType($1);
		}
	| _SPEC_STRG_CL
		{ $$ = new FSBaseType($1);
		}
;
type_specifier
	: TYPE
		{ $$ = new FSBaseType($1);
		  ((FSBaseType *)($$))->CheckTypeDef(typDfAllowed);
		}
   | const_volat
		{ $$ = new FSBaseType($1);
		  ((FSBaseType *)($$))->CheckTypeDef(typDfAllowed);
		}
   | _SBIT
		{ $$ = new FSBaseType($1);
		  ((FSBaseType *)($$))->CheckTypeDef(typDfAllowed);
		}
   | _BIT
		{ $$ = new FSBaseType($1);
		  ((FSBaseType *)($$))->CheckTypeDef(typDfAllowed);
		}
   | _SFR
		{ $$ = new FSBaseType($1);
		  ((FSBaseType *)($$))->CheckTypeDef(typDfAllowed);
		}
   | _SFRP
		{ $$ = new FSBaseType($1);
		  ((FSBaseType *)($$))->CheckTypeDef(typDfAllowed);
		}
   | _SFR16
		{ $$ = new FSBaseType($1);
		  ((FSBaseType *)($$))->CheckTypeDef(typDfAllowed);
		}
	| { typDfAllowed = FALSE; } enum_specifier
		{ $$ = $2;
		}
	| { typDfAllowed = FALSE; } struct_specifier
		{ $$ = $2;
		}
	| { typDfAllowed = FALSE; } TTYPE
		{ $$= $2->GetType()->Copy();
		  ((FSBaseType *)($$->GetBaseType()))->RemoveSpecifier(FS_TYPEDEF);
		}
   | _SEGMENT
      { $$ = new FSBaseType($1);
		  ((FSBaseType *)($$))->CheckTypeDef(typDfAllowed);
      }
;

	/* ------------------Enumerated Constants----------------------- */

enum_specifier
	: ENUM opt_enum_tag LC enumerator_list RC
		{ FSSymbol *oldSym = $2;
		  fsSymTable->InsertStrUnEnum(&oldSym);
		  if (oldSym == $2)
        {
		    ((FSStrUnEnumDefType *)(oldSym->GetType()))->SetMembers($4);
		    FSInsertSymListtoTable(*fsSymTable, *($4));
		    $$ = new FSStrUnEnumType(FS_T_ENUM,oldSym);
		  }
		  else
        {
           FSAnzError( E_MULT_DEF_s, oldSym->GetId() );
		     delete $4;
		    //YYERROR; already defined
		  }
		}
   | ENUM opt_enum_tag LC enumerator_list error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         FSSymbol *oldSym = $2;
		  fsSymTable->InsertStrUnEnum(&oldSym);
		  if (oldSym == $2) {
		    ((FSStrUnEnumDefType *)(oldSym->GetType()))->SetMembers($4);
		    FSInsertSymListtoTable(*fsSymTable, *($4));
		    $$ = new FSStrUnEnumType(FS_T_ENUM,oldSym);
		  }
		  else {
		    delete $4;
		    //YYERROR; already defined
		  }
		}
   | ENUM opt_enum_tag LC error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         FSSymbol *oldSym = $2;
		   fsSymTable->InsertStrUnEnum(&oldSym);
		   if (oldSym != $2)
         {
            FSAnzError( E_MULT_DEF_s, oldSym->GetId() );
         }
		    $$ = new FSStrUnEnumType(FS_T_ENUM,oldSym);
		}
	| ENUM NAME
		{ FSSymbol *pSym;
		  if ( fsSymTable->LookUpStrUnEnum( $2, &pSym ) == TRUE )
        {
            $$ = new FSStrUnEnumType( FS_T_ENUM, pSym );
		  }
		  else
        {
          pSym = new FSSymbol( $2, FS_NO_ID_COPY, new FSStrUnEnumDefType( FS_T_ENUM ) );
          fsSymTable->InsertStrUnEnum( &pSym );
          $$ = new FSStrUnEnumType( FS_T_ENUM, pSym );
		  }
		}
;
opt_enum_tag
	: NAME
		{ $$ = new FSSymbol($1,FS_NO_ID_COPY,new FSStrUnEnumDefType(FS_T_ENUM),FS_DEFINED);
		}
	| /* empty */
		{ $$ = new FSSymbol(FSUniqueName(),FS_ID_COPY,new FSStrUnEnumDefType(FS_T_ENUM),FS_DEFINED);
		}
;
enumerator_list
	: enumerator
        { $1->SetType( new FSStrUnEnumType(FS_T_ENUM,$<ptrSym>-1,FS_CONST) );
		  $$ = new FSSymbolList;
		  $$->Append($1);
		}
	| enumerator_list COMMA enumerator
        { /* $$ = $1 */
		  $3->SetType( new FSStrUnEnumType(FS_T_ENUM,$<ptrSym>-1,FS_CONST) );
		  $$->Append($3);
		}
;
enumerator
	: NAME
		{ $$ = new FSSymbol($1,FS_NO_ID_COPY,0,FS_DEFINED);
		}
	| NAME EQUAL const_expr
		{ $$ = new FSSymbol($1,FS_NO_ID_COPY,0,FS_DEFINED);
		}
   | NAME EQUAL error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $$ = new FSSymbol($1,FS_NO_ID_COPY,0,FS_DEFINED);
		}
;

	/* -------------------Variable Declarators---------------------- */

var_decl
	: decl_name		%prec COMMA /* $$ = $1 */
	| var_decl LP rp_c166
		{ $$ = FSAppendType(new FSFunctType(FS_S_KR), $1 );
		}
   | var_decl LP error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $$ = FSAppendType(new FSFunctType(FS_S_KR), $1 );
		}
	| var_decl LP var_list rp_c166
		{ $$ = FSAppendType(new FSFunctType(FS_S_ANSI,$3), $1 );
		}
   | var_decl LP var_list error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $$ = FSAppendType(new FSFunctType(FS_S_ANSI,$3), $1 );
		}
	| var_decl LB rb
		{ $$ = FSAppendType(new FSArrType(FS_NO_DIM_PRESENT), $1 );
		}
   | var_decl LB error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $$ = FSAppendType(new FSArrType(FS_NO_DIM_PRESENT), $1 );
		}
	| var_decl LB const_expr rb
		{ $$ = FSAppendType(new FSArrType(FS_DIM_PRESENT), $1 );
		}
   | var_decl LB const_expr error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $$ = FSAppendType(new FSArrType(FS_DIM_PRESENT), $1 );
		}
	| star var_decl		%prec UNOP
		{ $$ = FSAppendType(new FSPtrType($1), $2);
		}
	| lp_modif var_decl RP
		{ $$ = $2; }
   | lp_modif var_decl error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $$ = $2;
      }
;
rb
   : RB
   | rb _HW_SPEC_QUAL
;
decl_name
	: new_name         /* $$ = $1 */
	| lp_modif decl_name RP  /* nije redundantno - da bi moglo int (x)(...) za deklr. fje */
		{ $$ = $2;
		}
   /* lp_modif decl_name error nije potrebno jer se od lp_modif decl_name pravi var_decl */
;
new_name
	: NAME
		{ $$ = new FSSymbol( $1,FS_NO_ID_COPY );
          typDfAllowed = TRUE;
		}
;

lp_modif
   : LP
   | lp_modif spec_mod
   | lp_modif _HI_TECH_SPEC_QUAL
;
rp_c166
   : RP
   | RP _C166_REGMASK ICON
;
star
   : STAR { $$ = FS_NOTYPE; }
   | STAR const_volat_list { $$ = $2; }
;
const_volat_list
   : const_volat /* $$ = $1 */
   | spec_mod /* $$ = $1 */
   | const_volat_list const_volat { $$ = (FSBaseTypeSpecifiers)($1 | $2); }
   | const_volat_list spec_mod { $$ = (FSBaseTypeSpecifiers)($1 | $2); }
;
const_volat
   : CONST_VOLAT /* $$ = $1 */
   | _SPEC_QUAL /* $$ = $1 */
   | _HW_SPEC_QUAL /* $$ = $1 */
   | _HI_TECH_SPEC_QUAL /* $$ = $1 */
   | _IAR_INT  %prec COMMA /* $$ = $1 */
   | _IAR_INT LB const_expr RB /* $$ = $1 */
   | _IAR_INT _IAR_USING LB const_expr RB /* $$ = $1 */
   | _IAR_INT LB const_expr RB _IAR_USING LB const_expr RB /* $$ = $1 */
   | _COSMIC_6805_AT identifier_or_typedef_name { $$ = FS_NOTYPE; }
;
	/* --------------------Function Declarators--------------------- */

funct_decl
	: star funct_decl  %prec STAR
		{ $$ = FSAppendType(new FSPtrType($1),$2);
		}
	| funct_decl LB rb
		{ $$ = FSAppendType(new FSArrType(FS_NO_DIM_PRESENT),$1);
		}
   | funct_decl LB error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $$ = FSAppendType(new FSArrType(FS_NO_DIM_PRESENT),$1);
		}
	| funct_decl LB const_expr rb
		{ $$ = FSAppendType(new FSArrType(FS_DIM_PRESENT),$1);
		}
   | funct_decl LB const_expr error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $$ = FSAppendType(new FSArrType(FS_DIM_PRESENT),$1);
		}
	| lp_modif funct_decl RP
		{ $$ = $2; }
   | lp_modif funct_decl error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $$ = $2;
      }
	| funct_decl LP rp_c166
		{ $$ = FSAppendType(new FSFunctType(FS_S_KR),$1);
		}
   | funct_decl LP error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $$ = FSAppendType(new FSFunctType(FS_S_KR),$1);
		}
	| decl_name LP rp_c166
		{ $$ = FSAppendType(new FSFunctType(FS_S_KR),$1);
		}
   | decl_name LP error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $$ = FSAppendType(new FSFunctType(FS_S_KR),$1);
		}
	| decl_name LP name_list rp_c166
		{ $$ = FSAppendType(new FSFunctType(FS_S_KR,$3),$1);
		}
   | decl_name LP name_list error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $$ = FSAppendType(new FSFunctType(FS_S_KR,$3),$1);
		}
	| decl_name LP var_list rp_c166
		{ $$ = FSAppendType(new FSFunctType(FS_S_ANSI,$3),$1);
		}
   | decl_name LP var_list error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $$ = FSAppendType(new FSFunctType(FS_S_ANSI,$3),$1);
		}
;
name_list
	: new_name
		{ $$ = new FSSymbolList;
        $1->SetType( new FSBaseType( FS_INT ) );
		  $$->Append($1);
		}
	| name_list COMMA new_name
		{  /* $$ = $1 */
         $3->SetType( new FSBaseType( FS_INT ) );
         $$->Append($3);
		}
;

var_list
	: param_declaration
		{ $$ = new FSSymbolList;
		  if ( $1 ) $$->Append($1);
		}
	| var_list COMMA param_declaration
		{ /* $$ = $1 */ if ( $3 ) $$->Append($3);
		}
;
param_declaration
	: type var_decl
		{ $$ = FSAppendType( $1, $2 );
		}
   /* | type error not needed because abstract_decl is recognized in case of */
   /* error after type */
	| abstract_decl /* $$ = $1 */
	| ELLIPSIS
		{ /* treba uvesti neki flag u FunctType */
	      $$ = 0;
		}
;

	/* ---------------------Abstract Declarators--------------------- */

abstract_decl
	: type abs_decl
		{ $$ = FSAppendType( $1, $2 );
         typDfAllowed = TRUE;
		}
	| type %prec error
		{ $$ = FSAppendType( $1, new FSSymbol("(abstract)") );
          typDfAllowed = TRUE;
		}
	| type error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $$ = FSAppendType( $1, new FSSymbol("(abstract)") );
          typDfAllowed = TRUE;
		}
;
abs_decl
	: lp_modif abs_decl RP
		{
         $$ = $2;
		}
   | lp_modif abs_decl error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $$ = $2;
		}
	| star abs_decl %prec STAR
		{ $$ = FSAppendType(new FSPtrType($1), $2 );
		}
   | star %prec STAR
		{ $$ = FSAppendType(new FSPtrType($1), new FSSymbol("(abstract)") );
		}
	| abs_decl LB rb
		{ $$ = FSAppendType(new FSArrType(FS_NO_DIM_PRESENT),$1);
		}
   | abs_decl LB error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $$ = FSAppendType(new FSArrType(FS_NO_DIM_PRESENT),$1);
		}
	| abs_decl LB const_expr rb
		{ $$ = FSAppendType(new FSArrType(FS_DIM_PRESENT),$1);
		}
	| abs_decl LB const_expr error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $$ = FSAppendType(new FSArrType(FS_DIM_PRESENT),$1);
		}
	| LB rb
		{ $$ = FSAppendType(new FSArrType(FS_NO_DIM_PRESENT),new FSSymbol("(abstract)"));
		}
	| LB error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $$ = FSAppendType(new FSArrType(FS_NO_DIM_PRESENT),new FSSymbol("(abstract)"));
		}
	| LB const_expr rb
		{ $$ = FSAppendType(new FSArrType(FS_DIM_PRESENT),new FSSymbol("(abstract)"));
		}
	| LB const_expr error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $$ = FSAppendType(new FSArrType(FS_DIM_PRESENT),new FSSymbol("(abstract)"));
		}
	| abs_decl LP rp_c166
		{ $$ = FSAppendType(new FSFunctType(FS_S_KR), $1);
		}
	| abs_decl LP error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $$ = FSAppendType(new FSFunctType(FS_S_KR), $1);
		}
   | abs_decl LP var_list rp_c166
		{ $$ = FSAppendType(new FSFunctType( FS_S_ANSI, $3 ), $1);
		}
   | abs_decl LP var_list error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $$ = FSAppendType(new FSFunctType( FS_S_ANSI, $3 ), $1);
		}
   | LP rp_c166
		{ $$ = FSAppendType(new FSFunctType(FS_S_KR), new FSSymbol("(abstract)") );
		}
   | LP error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $$ = FSAppendType(new FSFunctType(FS_S_KR), new FSSymbol("(abstract)") );
		}
   | LP var_list rp_c166
		{ $$ = FSAppendType(new FSFunctType( FS_S_ANSI, $2 ), new FSSymbol("(abstract)") );
		}
   | LP var_list error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $$ = FSAppendType(new FSFunctType( FS_S_ANSI, $2 ), new FSSymbol("(abstract)") );
		}
;

	/* -------------------------Structures---------------------------- */

struct_specifier
	: STRUCT opt_tag LC struct_def_list { typDfAllowed = FALSE; } RC
		{
        if ($2->Defined() == FS_DEFINED )
        {
		    FSAnzError( E_MULT_DEF_s, $2->GetId() );
		    delete $4;
		  }
		  else
        {
            ((FSStrUnEnumDefType *)($2->GetType()))->SetMembers($4);
             $2->SetDef(FS_DEFINED);
        }
		  $$ = new FSStrUnEnumType($1,$2);
		}
	/* | STRUCT opt_tag LC error nije potrebno jer struct_def_list moze biti prazna */
	| STRUCT tag
		{  $$ = new FSStrUnEnumType($1,$2);
		}
;
opt_tag
	: NAME
		{ $$ = new FSSymbol($1,FS_NO_ID_COPY, new FSStrUnEnumDefType($<typId>0));
		  fsSymTable->InsertStrUnEnum(&($$));
		  typDfAllowed = TRUE; /* lookahead is already scanned, but it is { */
		}
	| /* empty */
		{ $$ = new FSSymbol(FSUniqueName(),FS_ID_COPY,new FSStrUnEnumDefType($<typId>0));
		  fsSymTable->InsertStrUnEnum(&($$));
		  typDfAllowed = TRUE; /* lookahead is already scanned, but it is { */
		}
;
tag
	: NAME
		{ if (fsSymTable->LookUpStrUnEnum($1,&($$)) == FALSE)
        {
		    /* install a new uncomplete type at the current level */
		    $$ = new FSSymbol($1,FS_NO_ID_COPY,new FSStrUnEnumDefType($<typId>0));
		    fsSymTable->InsertStrUnEnum(&$$);
		  }
        else
        {
         delete [] $1;
        }
		}
;
	/* -------------- Structure & Union members ---------------- */
struct_def_list
	: struct_def_list struct_def /* $$ = $1 */
	| /* empty */
		{
         $$ = new FSSymbolList;
		}
;
struct_def
//	: error st_err_end
   /* greska na pocetku se hvata u struct_specifier */
	: opt_specifiers struct_decl_list SEMI /* no action */
   | opt_specifiers struct_decl_list error
      {
         FSAnzError( E_INV_DECL );
         yyerrok;
      }
	| opt_specifiers { typDfAllowed = TRUE; } SEMI
		{ delete $1;
		  FSAnzError( E_INV_STRUCT_ELEM );
		}
   | opt_specifiers error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         if ( pcyytoken == NAME ) delete [] yylval.ptrChar;
         while ( ( pcyytoken = yylex() ) != SEMI && pcyytoken != RC && pcyytoken != EOF )
         {
            if ( pcyytoken == NAME ) delete [] yylval.ptrChar;
         }
         if ( pcyytoken == SEMI ) pcyytoken = yylex();
         if ( pcyytoken == EOF ) pcyytoken = 0;
         delete $1;
		}
;
struct_decl_list
	: struct_decl
		{ $1 = FSAppendType($<ptrType>0, $1, (yychar == COMMA ) ? typDfAllowed = FALSE, FS_SPEC_COPY : FS_NO_SPEC_COPY );
        if ( yychar != COMMA ) $<ptrType>0 = 0;
		  $<ptrSymList>-1 ->Append( $1 );
		}
	| struct_decl_list comma_modif struct_decl
		{ $3 = FSAppendType($<ptrType>0, $3, (yychar == COMMA ) ? typDfAllowed = FALSE, FS_SPEC_COPY : FS_NO_SPEC_COPY );
         if ( yychar != COMMA ) $<ptrType>0 = 0;
		  $<ptrSymList>-1 ->Append( $3 );
		}
;
struct_decl
	: var_decl   /* $$ = $1 */
	| var_decl COLON const_expr /* %prec COMMA *//* $$ = $1 */
   | var_decl COLON error
      { /* $$ = $1 */
         FSAnzError( E_INV_STRUCT_ELEM );
         yyerrok;
      }
	| COLON const_expr /* %prec COMMA */
      {
         $$ = new FSSymbol(FSUniqueName(), FS_ID_COPY, new FSBaseType( FS_INT ) );
         typDfAllowed = TRUE;
      }

;
	/* --------------------K&R Function Arguments--------------------- */

arg_def_list
	: arg_def_list arg_def
	| c51_func_attr
;
c51_func_attr
    : /* Empty */
    | c51_func_attr _C51_NOPAR
    | c51_func_attr _PRIORITY ICON
    | c51_func_attr _USING identifier_or_typedef_name
    | c51_func_attr _USING ICON
    | c51_func_attr _TASK ICON
    | c51_func_attr _C51_166_INT identifier_or_typedef_name
    | c51_func_attr _C51_166_INT ICON
    | c51_func_attr _C51_166_INT identifier_or_typedef_name EQUAL ICON
;
identifier_or_typedef_name
   : NAME { delete [] $1; }
   | TTYPE
;

arg_def
	: specifiers error /* error arg_err_end */
      {
         FSAnzError( E_INV_PARAM_DECL );
         yyerrok;
         if ( pcyytoken == NAME ) delete [] yylval.ptrChar;
         while ( ( pcyytoken = yylex() ) != SEMI && pcyytoken != EOF )
         {
            if ( pcyytoken == NAME ) delete [] yylval.ptrChar;
         }
         if ( pcyytoken == SEMI ) pcyytoken = yylex();
         if ( pcyytoken == EOF ) pcyytoken = 0;
         delete $1;
      }
	| specifiers arg_decl_list SEMI
	| specifiers { typDfAllowed = TRUE; } SEMI
		{
		  FSTypeId type = $1->GetId();
        if ( type != FS_T_STRUCT && type != FS_T_UNION && type != FS_T_ENUM )
            FSAnzError( E_INV_PARAM_DECL );
        delete $1;
		}
;
arg_decl_list
	: arg_decl
		{ $1 = FSAppendType($<ptrType>0, $1, (yychar == COMMA ) ? typDfAllowed = FALSE, FS_SPEC_COPY : FS_NO_SPEC_COPY );
         if ( yychar != COMMA ) $<ptrType>0 = 0;
		  FSPatchArgs($<ptrSym>-2, $1 );
		}
	| arg_decl_list comma_modif arg_decl
		{ $3 = FSAppendType($<ptrType>0, $3, (yychar == COMMA ) ? typDfAllowed = FALSE, FS_SPEC_COPY : FS_NO_SPEC_COPY );
         if ( yychar != COMMA ) $<ptrType>0 = 0;
		  FSPatchArgs( $<ptrSym>-2, $3 );
		}
;
arg_decl
	: funct_decl /* $$ = $1 */
	| var_decl   /* $$ = $1 */
;

	/* ------------Local Variables ---------------------------------- */

def_list
	: def_list def
	| /* empty */
;
def
	: specifiers error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         if ( pcyytoken == NAME ) delete [] yylval.ptrChar;
         while ( ( pcyytoken = yylex() ) != SEMI && pcyytoken != EOF )
         {
            if ( pcyytoken == NAME ) delete [] yylval.ptrChar;
         }
         if ( pcyytoken == SEMI ) pcyytoken = yylex();
         if ( pcyytoken == EOF ) pcyytoken = 0;
         delete $1;
      }
	| specifiers decl_list SEMI
	| specifiers { typDfAllowed = TRUE; } SEMI
		{
        FSTypeId type = $1->GetId();
		  if ( type != FS_T_STRUCT && type != FS_T_UNION && type != FS_T_ENUM )
            FSAnzError( E_INV_DECL );
        delete $1;
		}
;
decl_list
	: decl
		{ $1 = FSAppendType($<ptrType>0, $1, (yychar == COMMA ) ? typDfAllowed = FALSE, FS_SPEC_COPY : FS_NO_SPEC_COPY );
         if ( yychar != COMMA ) $<ptrType>0 = 0;
		  fsSymTable->Insert( $1 );
		}
	| decl_list comma_modif decl
		{ $3 = FSAppendType($<ptrType>0, $3, (yychar == COMMA ) ? typDfAllowed = FALSE, FS_SPEC_COPY : FS_NO_SPEC_COPY );
         if ( yychar != COMMA ) $<ptrType>0 = 0;
		  fsSymTable->Insert( $3 );
		}
;
decl
	: funct_decl /* $$ = $1 */
	| var_decl   /* $$ = $1 */
	| var_decl EQUAL initializer
		{ $1->SetDef(FS_DEFINED);
		}
   | var_decl EQUAL error
		{
         FSAnzError( E_INV_DECL );
         yyerrok;
         $1->SetDef(FS_DEFINED);
		}
;
initializer
	: expr
	| LC initializer_list RC /* %prec COMMA */
;
initializer_list
	: initializer
	| initializer_list COMMA initializer
;

	/* ----------------------Statements--------------------------- */

compound_stmt
	: LC local_defs stmt_list RC
;
local_defs
	: def_list
;
stmt_list
	: stmt_list statement
	| /* empty */
;
statement
	: error
      {
         FSAnzError( E_INV_STMT );
         yyerrok;
         if ( pcyytoken == NAME ) delete [] yylval.ptrChar;
         while (    ( pcyytoken = yylex() ) != SEMI
                 && pcyytoken != EOF
                 && pcyytoken != LC
                 && pcyytoken != RC
               )
         {
            if ( pcyytoken == NAME ) delete [] yylval.ptrChar;
         }
         if ( pcyytoken == SEMI ) pcyytoken = yylex();
         if ( pcyytoken == EOF ) pcyytoken = 0;
      }
	| expr SEMI
	|	{fsSymTable->EnterScope();
		}
		compound_stmt
		{fsSymTable->LeaveScope();
		}
	| RETURN SEMI
	| RETURN expr SEMI
	| SEMI
	| GOTO target SEMI
	| target COLON {} statement
	| SWITCH LP expr RP
		{ fsSymTable->EnterScope();
		}
		compound_stmt
		{ fsSymTable->LeaveScope();
		}
	| CASE const_expr COLON
	| DEFAULT COLON
	| IF LP test RP statement %prec ELSE
	| IF LP test RP statement ELSE {} statement
	| WHILE LP test RP {} statement
	| FOR LP opt_expr SEMI test SEMI {} opt_expr RP {} statement
	| DO {} statement WHILE {} LP test RP SEMI
	| BREAK SEMI
	| CONTINUE SEMI
   | _TRY compound_stmt
   | _FINALLY compound_stmt
   | _EXCEPT LP expr RP compound_stmt
   | _ASM
   | _HI_TECH_ASM LP string_const RP SEMI
;
test
	: {} expr
	| /* empty */
;
target
	: NAME  { delete [] $1; }
;

	/* ------------------------Expressions------------------------- */

opt_expr
	: expr
	| /* empty */
;
const_expr
	: expr  %prec COMMA
;
expr
	: expr COMMA {} non_comma_expr
	| non_comma_expr
;
non_comma_expr
	: non_comma_expr QUEST {} non_comma_expr COLON {} non_comma_expr
	| non_comma_expr ASSIGNOP non_comma_expr
	| non_comma_expr EQUAL non_comma_expr
	| or_expr
;
or_expr
	: or_list
;
or_list
	: or_list OROR {} and_expr
	| and_expr
;
and_expr
	: and_list
;
and_list
	: and_list ANDAND {} binary
	| binary
;
binary
	: binary RELOP binary
	| binary EQUOP binary
	| binary STAR binary
	| binary DIVOP binary
	| binary SHIFTOP binary
	| binary AND binary
	| binary XOR binary
	| binary OR binary
	| binary PLUS binary
	| binary MINUS binary
	| unary
;
unary
	: LP expr RP
	| FCON
   | _IAR_CON
	| ICON
   | _SELF
   | _IAR_ARGS
   | _IAR_UNARY LP expr RP
   | unary _BASE_OPERATOR unary
   | _SEGNAME LP string_const RP
	| NAME
		{
         FSSymbol *pSym;
         if ( fsSymTable->LookUp( $1, &pSym ) )
         {
             /* pSym->GetLevel() == FS_EXT_DECL za globalne simbole, ostalo su lokalni */
             /* pSym->FirstUseInCurrFunc() == TRUE za prvu upotrebu, FALSE za sledece */
             pSym->SetFirstUseInCurrFunc( FALSE );
         }
         delete [] $1;
		}
	| string_const  %prec COMMA
	| SIZEOF LP string_const RP  %prec SIZEOF
	| SIZEOF LP expr RP  %prec SIZEOF
	| SIZEOF LP abstract_decl RP  %prec SIZEOF
      {
         delete $3;
      }
   | SIZEOF LP abstract_decl error
      {
         FSAnzError( E_INV_EXPR );
         yyerrok;
         delete $3;
      }
	| LP abstract_decl RP unary  %prec UNOP
      {
         delete $2;
      }
   | LP abstract_decl error
      {
         FSAnzError( E_INV_EXPR );
         yyerrok;
         delete $2;
      }
	| MINUS unary  %prec UNOP
	| UNOP unary
	| unary INCOP
	| INCOP unary
	| AND unary  %prec UNOP
	| STAR unary  %prec UNOP
	| unary LB expr RB  %prec UNOP
	| unary STRUCTOP NAME  %prec STRUCTOP { delete [] $3; }
   | unary STRUCTOP _IAR_CON
   | unary _IAR_CON
	| unary LP args RP
	| unary LP RP
;
string_const
	: STRING
	| string_const STRING
;
args
	: non_comma_expr  %prec COMMA
	| non_comma_expr COMMA args
;
%%