//----------------------------------------------------------------------------
//  Description :
//      preprocessor control class FSPreProc implementation
//----------------------------------------------------------------------------
//  Changes:
//  $Log__$
//----------------------------------------------------------------------------

#include <string.h>
#include <fstream.h>
#include <stdlib.h>
//#include <dir.h>
#ifdef PPDEBUG
   #include <iostream.h>
#endif

#ifdef DEBUG
   #ifdef BRANKO
      #include "debug.h"
   #endif
#endif

#include "fsbool.h"
#include "fsanzopt.h"
#include "fserror.h"
#include "fsifexpr.h"
#include "fspproc.h"
#include "fsstacks.h"
#include "fstoken.h"
//#include "fssymtab.h" // fspars_t.h needs it
#include "fspars_t.h"

//char *yytext;
//int yyleng;

static const FS_MAX_INCLUDE_ARG_LEN = 80;

FSPreProc *preProc;

//##ModelId=419A04D601E1
enum FSPreProcKeyword
{
	//##ModelId=419A04D601F6
   FSK_DEFINE,
	//##ModelId=419A04D601FF
	FSK_UNDEF,
	//##ModelId=419A04D60209
   FSK_INCLUDE,
	//##ModelId=419A04D60213
   FSK_PRAGMA,
	//##ModelId=419A04D6021D
   FSK_ERROR,
	//##ModelId=419A04D60227
	FSK_IF,
	//##ModelId=419A04D60231
   FSK_IFDEF,
	//##ModelId=419A04D6023B
   FSK_IFNDEF,
	//##ModelId=419A04D60245
   FSK_ELSE,
	//##ModelId=419A04D60263
   FSK_ELIF,
	//##ModelId=419A04D6026D
   FSK_ENDIF,
	//##ModelId=419A04D60278
   FSK_NULL,
	//##ModelId=419A04D60282
   FSK_ASM,
	//##ModelId=419A04D6028C
   FSK_ENDASM,
	//##ModelId=419A04D60296
   FSK_OTHER
};

static FSPreProcKeyword FSPreProcKeywordId( char *str )
{
   switch ( *str )
   {
		case '\n':
           return FSK_NULL;
      case 'a':
			  if (    strcmp( str, "asm" ) == 0
                && ( fsAnlzOptions.GetMode() & FSM_HI_TECH_6805 )
              ) return FSK_ASM;
           break;
      // case EOF impossible because \n must precede it when reading from file
      case 'd':
           if ( strcmp( str, "define" ) == 0 ) return FSK_DEFINE;
           break;
      case 'i':
           if ( strcmp( str, "include" ) == 0 ) return FSK_INCLUDE;
           else if ( strcmp( str, "if" ) == 0 ) return FSK_IF;
			  else if ( strcmp( str, "ifdef" ) == 0 ) return FSK_IFDEF;
           else if ( strcmp( str, "ifndef" ) == 0 ) return FSK_IFNDEF;
           break;
      case 'u':
			  if ( strcmp( str, "undef" ) == 0 ) return FSK_UNDEF;
           break;
      case 'e':
           if ( strcmp( str, "else" ) == 0 ) return FSK_ELSE;
			  else if ( strcmp( str, "endif" ) == 0 ) return FSK_ENDIF;
           else if ( strcmp( str, "elif" ) == 0 ) return FSK_ELIF;
			  else if ( strcmp( str, "error" ) == 0 ) return FSK_ERROR;
			  else if (    strcmp( str, "endasm" ) == 0
                     && ( fsAnlzOptions.GetMode() & FSM_HI_TECH_6805 )
                   ) return FSK_ENDASM;
			  break;
		case 'p':
			  if ( strcmp( str, "pragma" ) == 0 ) return FSK_PRAGMA;
			  break;
	}
	return FSK_OTHER;
}

//----------------------------------------------------------------------------
// Function    : NotFullDir
//----------------------------------------------------------------------------
// Description : Checks that file or directory name is full
//
// Parameters  : ADir - string to check
//
// Returns     : FALSE if ADir has ":\" as 2. and 3. characters
//----------------------------------------------------------------------------
static Boolean NotFullDir( char* ADir )
{
   return ( Boolean )( strlen(ADir) < 3 || ADir[1] != ':' || ADir[2] != '\\');
}

//----------------------------------------------------------------------------
// Function    : MakeAbsPathname
//----------------------------------------------------------------------------
// Description : Joins relative file or directory name in AFname to
//               absolute directory in ADir
//
// Parameters  : ADst - destination buffer
//               ADir- directory ends with last "\", text after that
//               character is ignored
//               AFname - filename to complete
//
// Returns     : none
//----------------------------------------------------------------------------
static void MakeAbsPathname( char* ADst, char* ADir, char* AFname )
{
   ADst[_MAX_PATH - 1] = '\0';

   if ( NotFullDir( ADir ) || strlen(AFname) < 1 ||
        (AFname[0] != ADir[0] && AFname[1] == ':') )
   {
      strncpy(ADst, AFname, _MAX_PATH - 1);
      return;
   }

   char* d = AFname;

   if ( d[1] == ':' )
      d += 2;

   while (strncmp(d, ".\\", 2) == 0)
      d += 2;

   strncpy(ADst, ADir, _MAX_PATH - 1);

   if (d[0] == '\0' || strcmp(d, ".") == 0)
      return;

   if (d[0] == '\\')
   {
      strncpy(ADst + 2, d, _MAX_PATH - 3);
      return;
   }

   char* p = strrchr(ADst, '\\');
   while ( p != NULL && strncmp(d, "..\\", 3) == 0)
   {
      *p = '\0';
      p = strrchr(ADst, '\\');
      d += 3;
      while (strncmp(d, ".\\", 2) == 0)
         d += 2;
   }

   if ( p != NULL )
      strncpy( p + 1, d, _MAX_PATH - 2 - ( p - ADst ) );
   else
      strncpy(ADst, AFname, _MAX_PATH - 1 );
}


//##ModelId=419A04D501FF
int FSPreProc::GetToken()
{
	int token;
	Boolean lineBeg;
	while ( 1 )
	{
		FSMacro *pMacro;
   	if ( this->tokenIsSaved )
	   {
		   this->tokenIsSaved = FALSE;
		   token = this->savedToken;
         this->outBuffer.RedirectToInternalBuffer();
	   }
      else
      {
		   token = this->tokenizer.Get();
      }
		switch( token )
		{
			case SHARP:
				lineBeg = this->lineBeginning;
				this->lineBeginning = FALSE;
				if ( lineBeg && this->macroCallStack.IsNotEmpty() == FALSE )
				{
					 // preprocessor directive
					 token = this->tokenizer.Get();
					 FSPreProcKeyword preProcKeyword;
					 switch ( preProcKeyword = FSPreProcKeywordId ( this->outBuffer.GetTxt() ) )
					 {
                   case FSK_PRAGMA:
                   case FSK_ASM:
                   case FSK_ENDASM:
						 case FSK_NULL:
								break;

						 case FSK_UNDEF:
								if ( ( token = this->tokenizer.Get() ) != NAME )
								{
									FSAnzError( E_PP_INV_MACRO_NAME );
								}
								else
								{ // macro definition
									this->macroTable.Remove( this->outBuffer.GetTxt() );
									token = this->tokenizer.Get();
								}
								break;

						 case FSK_DEFINE:
								if ( ( token = this->tokenizer.Get() ) != NAME )
								{
                           FSAnzError( E_PP_INV_MACRO_NAME );
								}
								else
								{ // macro definition
									this->macroTable.Insert( this->outBuffer.GetTxt(), token );
								}
								break;

						 case FSK_IF:
						 case FSK_IFDEF:
						 case FSK_IFNDEF:
						 case FSK_ELSE:
                   case FSK_ELIF:
						 case FSK_ENDIF:
								while ( 1 )
								{
                           Boolean skipTxt = ( Boolean )( this->ifStack.IsNotEmpty() && this->ifStack.Top().skipText );
									switch( preProcKeyword )
									{
										case FSK_IFDEF:
                                 if ( skipTxt )
                                 {
                                    this->ifStack.Push( TRUE, TRUE, this->Line() );
                                 }
                                 else if ( ( token = this->tokenizer.Get() ) == NAME )
                                 {
                                    this->ifStack.Push(
                                              FALSE,
                                              ( ( this->macroTable.LookUp( this->outBuffer.GetTxt() ) != 0 )
                                                 ? FALSE
                                                 : TRUE
                                              ),
                                              this->Line()
                                    );
                                 }
                                 else FSAnzError( E_PP_INV_X_DIR_s, "ifdef" );
											break;

										case FSK_IFNDEF:
                                 if ( skipTxt )
                                 {
                                    this->ifStack.Push( TRUE, TRUE, this->Line() );
                                 }
                                 else if ( ( token = this->tokenizer.Get() ) == NAME )
                                 {
                                    this->ifStack.Push (
                                              FALSE,
                                              ( ( this->macroTable.LookUp( this->outBuffer.GetTxt() ) != 0 )
                                               ? TRUE
                                               : FALSE
                                              ),
                                              this->Line()
                                    );
                                 }
                                 else FSAnzError( E_PP_INV_X_DIR_s, "ifndef" );
											break;

                              case FSK_ELIF:
										case FSK_ELSE:
                                 {
                                    FSIf *iF;
                                    if ( this->ifStack.IsNotEmpty() && ( iF = &( this->ifStack.Top() ) )->elsePart == FALSE )
                                    {
                                       iF->elsePart = ( Boolean ) !iF->elsePart;
                                       iF->skipText = ( Boolean ) ( !iF->skipText || iF->ignore );
                                    }
                                    else
                                    {
                                       FSAnzError( E_PP_MISP_X_DIR_s, ( preProcKeyword == FSK_ELIF ) ? "elif" : "else" );
                                       break;
                                    }
											   if ( preProcKeyword == FSK_ELSE ) break;
                                    // FSK_ELIF
											   this->ifStack.Pop();
                                    skipTxt = ( Boolean )( this->ifStack.IsNotEmpty() && this->ifStack.Top().skipText );
                                    // goto case FSK_IF
                                 }

                              case FSK_IF:
                                 if ( skipTxt )
                                 {
                                    this->ifStack.Push( TRUE, TRUE, this->Line() );
                                 }
                                 else
                                 {
                                    this->ifStack.Push ( FALSE, ParseIfExpr( token ), this->Line() );
                                 }
                                 break;

										case FSK_ENDIF:
                                 if ( this->ifStack.IsNotEmpty() )
                                 {
                                    this->ifStack.Pop();
                                 }
                                 else FSAnzError( E_PP_MISP_X_DIR_s, "endif" );
											break;

										// default: no action

									}
                           if (    this->ifStack.IsNotEmpty() == FALSE
                                || this->ifStack.Top().skipText == FALSE
                              )
                                 break;
									// skipping the text until the next preproc. directive
                           while ( TRUE )
                           {
                              if ( token == NEWLINE )
                              {
                                 token = this->tokenizer.Get();
                                 if ( token == SHARP ) break;
                              }
                              else if ( token == EOF )
                                 break;
                              else
                                 token = this->tokenizer.Get();
                           }
                           if ( token == EOF )
                           {
                              break;
                           }
                           // else token == SHARP
                           token = this->tokenizer.Get();
                           preProcKeyword = FSPreProcKeywordId ( this->outBuffer.GetTxt() );
								}
								break;

						 case FSK_INCLUDE:
								char includeArg[ FS_MAX_INCLUDE_ARG_LEN + 1 ];
								includeArg[ FS_MAX_INCLUDE_ARG_LEN ] = '\0';
								*includeArg = '\0';
								this->tokenizer.processBackslashInString = FALSE;
								// macro expand the rest of the line
								while( 1 )
								{
									switch( token = this->tokenizer.Get() )
									{
										case NAME:
											// possible macro expansion
											if (     ( pMacro = this->macroTable.LookUp( this->outBuffer.GetTxt() ) ) != 0
													&& this->macroCallStack.ExpandingNow( pMacro ) == FALSE
												)
											{
												// there is a macro with that name & it is not recursive call
												if ( pMacro->Args().Parenth() == FALSE )  // macro need no parenthesis
												{
													this->macroCallStack.Push( *pMacro );
													continue;
												}
												else if ( ( token = this->tokenizer.Get() ) == LP )// macro with parenthesis & there are parenth. following a name
												{
													this->macroCallStack.Push( *pMacro );
													continue;
												}
												//else -> not a macro call -> default action after sw.
											}
											break;

										case EOF:
                                 if ( this->macroCallStack.IsNotEmpty() )
                                 {
											   this->macroCallStack.Pop();
											   continue;
                                 }
                                 else
                                 {
                                    goto IncludeExit;
                                 }

										case NEWLINE:
											if ( this->macroCallStack.IsNotEmpty() )
											{
                                    continue;
                                 }
                                 else
                                 {
												goto IncludeExit;
											}

										// no need for default -> actions done after switch
									}
									//default
									strncat( includeArg, this->outBuffer.GetTxt(), FS_MAX_INCLUDE_ARG_LEN - strlen( includeArg ) );
								}
						 IncludeExit:
								this->tokenizer.processBackslashInString = TRUE;
								// now open the file - !!! treba naknadno doterati za listu dir. i trazenje za " ako nema u tekucem
								{
									int argLen = strlen( includeArg );
									if ( argLen > 2 )
									{
										includeArg[ argLen - 1 ] = '\0'; // eliminate trailing > or "
                              char filName[ _MAX_PATH ];
                              MakeAbsPathname( filName, ( includeArg[0] == '<' ) ? fsAnlzOptions.GetStdLibIncludeDir() : "" , includeArg + 1 );
										this->fileStack.PushFile( filName );
									}
								}
								break;

						 case FSK_OTHER:
                      FSAnzError( E_PP_UNKNOWN_DIR );
							 break;
						 // default: bug
					 }
					 // skip to end of line ignoring the rest of the directive
					 while ( token != NEWLINE && token != EOF )
							 token = this->tokenizer.Get();
					 this->lineBeginning = TRUE;
                if ( preProcKeyword == FSK_ASM ) return _ASM_START;
                if ( preProcKeyword == FSK_ENDASM ) return _ASM_END;
				} // end if (preprocessor directive)...
				break;

			case NEWLINE:
				this->lineBeginning = TRUE;
            if ( passThroughNewline ) return NEWLINE;
				break;

			case SHARPSHARP: // greska 'illegal character'; ne u DA-C analizi
				this->lineBeginning = FALSE;
				break;

			case EOF:
				if ( this->macroCallStack.IsNotEmpty() )
				{
					this->macroCallStack.Pop();
					break;
				}
				else if ( this->fileStack.IsNotEmpty() )
				{
               if ( this->ifStack.IsNotEmpty() )
               {
                   // error: eof before #endif
                  FSAnzError( E_PP_UNTERM_COND_ST_u, this->ifStack.Top().startLine );
                  this->ifStack.Clear();
               }
					this->fileStack.Pop();
					break;
				}
				else return EOF;

			case NAME:
				this->lineBeginning = FALSE;
				if (     ( pMacro = this->macroTable.LookUp( this->outBuffer.GetTxt() ) ) != 0
						&& this->macroCallStack.ExpandingNow( pMacro ) == FALSE
					)
				{
					// there is a macro with that name & it is not recursive call
					if ( pMacro->Args().Parenth() == FALSE )  // macro need no parenthesis
					{
						this->macroCallStack.Push( *pMacro );
						break;
					}
					else if ( ( token = this->tokenizer.Get() ) == LP )// macro with parenthesis & there are parenth. following a name
					{
						this->macroCallStack.Push( *pMacro );
						break;
					}
					else // not a macro call
					{
						this->savedToken = token;
						this->outBuffer.RedirectToString( pMacro->Name() );
						this->tokenIsSaved = TRUE;
					}
				}
				// token is not a macro name
				return NAME; //IdOrKeyword( this->outBuffer.GetTxt() );

			default:
				this->lineBeginning = FALSE;
				return token;
		}
	}
}

#pragma warning( disable : 4355 )
//##ModelId=419A04D50208
FSPreProc::FSPreProc( char *fileName )
	: cppEnabled ( fsAnlzOptions.GetCppComments() ),
	  nestedComments( fsAnlzOptions.GetNestedComments() ),
     fileStack ( *( new FSFileStack( *this ) ) ),
     lineBeginning( TRUE ),
     tokenIsSaved( FALSE ),
     passThroughNewline( TRUE ),
     line( 0 ),
     file( "" ),
     macroTable ( *( new FSMacroTable( *this ) ) ),
     macroCallStack ( *( new FSMacroCallStack( *this ) ) ),
	  inputBuffer ( *( new FSInputBuffer( fileStack, macroCallStack ) ) ),
     outBuffer ( *( new FSOutBuffer() ) ),
     tokenizer ( *( new FSTokenizer( *this, inputBuffer, outBuffer ) ) ),
     ifStack( *( new FSIfStack ) )
{
   ::preProc = this;

   // insert predefined macros
   FSMacro *pMacro = new FSMacro( "__STDC__" );
   FSMacroBody::FSMacroBodyElem *pElem = pMacro->Body().FirstChunk();
   FSMacroBody::FSMacroBodyElem::Insert( pElem, "1", strlen("1") );
   macroTable.Insert( pMacro );

   pMacro = new FSMacro( "__FILE__" );
   pElem = pMacro->Body().FirstChunk();
   FSMacroBody::FSMacroBodyElem::Insert( pElem, 0, FS_AR_FILNAME );
   macroTable.Insert( pMacro );

   pMacro = new FSMacro( "__LINE__" );
   pElem = pMacro->Body().FirstChunk();
   FSMacroBody::FSMacroBodyElem::Insert( pElem, 0, FS_AR_LINENUM );
   macroTable.Insert( pMacro );

   // push .c file onto stack
	fileStack.PushFile( fileName );
}

//##ModelId=419A04D50212
FSPreProc::~FSPreProc()
{
   delete &ifStack;
   delete &tokenizer;
   delete &outBuffer;
   delete &inputBuffer;
	delete &macroCallStack;
   delete &macroTable;
   delete &fileStack;
}

