//----------------------------------------------------------------------------
//  Description :
//      preprocessor input & output buffers and
//      tokenizer implementation
//----------------------------------------------------------------------------
//  Changes:
//  $Log__$
//----------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fstream.h>

#ifdef DEBUG
   #ifdef BRANKO
      #include "debug.h"
   #endif
#endif

#include "fsbool.h"
#include "fsanzopt.h"
#include "fserror.h"
#include "fspproc.h"
#include "fsstacks.h"
#include "fstoken.h"
//#include "fssymtab.h" // fspars_t.h needs it
#include "fspars_t.h"

//----------------------------------------------------------------------
//  class FSOutBuffer implementation
//----------------------------------------------------------------------
//##ModelId=419A04D2027E
char FSOutBuffer::outBuffer[] = { 0 };
//##ModelId=419A04D20236
char *FSOutBuffer::yytext = 0;
//##ModelId=419A04D20240
int FSOutBuffer::yyleng = 0;
//----------------------------------------------------------------------
//  class FSTokenizer implementation
//----------------------------------------------------------------------

/* static int chr = ' ';
inline int FsGetChar()
{  register int c = chr;
	chr = getchar();
	if ( c == '\\' && chr =='\n' ) c = getchar(), chr = getchar();
	return c;
}
*/
#define FsGetChar() inputBuffer.GetChar()

#define NEXT do { outBuffer.Put( ( char )nextChar ); nextChar = FsGetChar(); } while( 0 )
#define CLEAR outBuffer.Clear()
#define LAST outBuffer.SetTerminator()
#define SKIP nextChar = FsGetChar()
#define OUT( c ) outBuffer.Put( c )

#define LETTER_EXCEPT_L '_':\
        case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':\
        case 'G': case 'H': case 'I': case 'J': case 'K': case 'M':\
        case 'N': case 'O': case 'P': case 'Q': case 'R': case 'S':\
        case 'T': case 'U': case 'V': case 'W': case 'X': case 'Y':\
        case 'Z':\
        case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':\
        case 'g': case 'h': case 'i': case 'j': case 'k': case 'l':\
        case 'm': case 'n': case 'o': case 'p': case 'q': case 'r':\
        case 's': case 't': case 'u': case 'v': case 'w': case 'x':\
        case 'y': case 'z'

#define LETTER LETTER_EXCEPT_L: case 'L'

#define DIGIT1TO9      '1': case '2': case '3': case '4': case '5':\
                  case '6': case '7': case '8': case '9'

#define DIGIT0TO7      '0': case '1': case '2': case '3': case '4':\
                  case '5': case '6': case '7'

#define DIGIT DIGIT1TO9: case '0'

#define HEXDIGIT DIGIT: case 'A': case 'a': case 'B': case 'b': case 'C': case 'c':\
                        case 'D': case 'd': case 'E': case 'e': case 'F': case 'f'

#define WHITE_SPACE       ' ': case '\t': case '\v': case '\f':\
                    case '\r': case '\b': case '\a'

//##ModelId=419A04D202DB
int FSTokenizer::Get()
{
//------------------------------------------------------------------
//        Finite state automaton
//------------------------------------------------------------------
   Start:
            CLEAR;
            // update line & file info for current token
            if ( preProc.fileStack.IsNotEmpty() )
            {
               preProc.File() = preProc.fileStack.Top().Name();
               preProc.Line() = preProc.fileStack.Top().Line();
            }
            else
            {
               preProc.File() = "";
               preProc.Line() = 0;
            }

            switch ( nextChar )
            {
               case LETTER_EXCEPT_L: NEXT; goto Identifier;

               case 'L':
                  NEXT; goto MayBeLongLiteral;

               case '$':
                  NEXT;
                  if ( fsAnlzOptions.GetMode() & FSM_IAR ) goto Identifier;
                  goto FoundIllegalChar;

					case '0': NEXT; goto Zero;

               case DIGIT1TO9: NEXT; goto DecimalNumber;

               case '\'': NEXT; goto CharLiteral;

               case '\"': NEXT; goto String;

               case '#': NEXT; goto Sharp;

               case WHITE_SPACE: NEXT; goto WhiteSpace;

               case '\n': NEXT; goto FoundNEWLINE;
//                  if ( mode == FS_PlainText ) yylineno++; // prebaciti u citanje

//             case '\\': NEXT; goto BackSlash; // prebaciti u citanje

               case '/': NEXT; goto Slash;

					case '(': NEXT; goto FoundLP;

					case ')': NEXT; goto FoundRP;

               case '[': NEXT; goto FoundLB;

               case ']': NEXT; goto FoundRB;

               case '{': NEXT; goto FoundLC;

               case '}': NEXT; goto FoundRC;

               case '?': NEXT; goto FoundQUEST;

               case ':': NEXT; goto Colon;

               case ',': NEXT; goto FoundCOMMA;

					case ';': NEXT; goto Semi;

               case '=':
                  NEXT; if ( nextChar == '=' ) { NEXT; goto FoundEQUOP; }
                        else goto FoundEQUAL;

               case '+':
						NEXT; if ( nextChar == '=' ) { NEXT; goto FoundASSIGNOP; }
                        else if ( nextChar == '+' ) { NEXT; goto FoundINCOP; }
                        else if (   nextChar == '?'
                                 && fsAnlzOptions.GetMode() & ( FSM_HW_HC16 | FSM_HW_851 | FSM_HW_XA | FSM_HW_HC5_11_12 )
                                ) NEXT;
                        goto FoundPLUS;

               case '-':
                  NEXT; if ( nextChar == '=' ) { NEXT; goto FoundASSIGNOP; }
                        else if ( nextChar == '-' ) { NEXT; goto FoundINCOP; }
                        else if ( nextChar == '>' ) { NEXT; goto FoundSTRUCTOP; }
                        else if (   nextChar == '?'
                                 && fsAnlzOptions.GetMode() & ( FSM_HW_HC16 | FSM_HW_851 | FSM_HW_XA | FSM_HW_HC5_11_12 )
                                ) NEXT;
                        goto FoundMINUS;

               case '.': NEXT; goto Dot;

               case '*':
                  NEXT; if ( nextChar == '=' ) { NEXT; goto FoundASSIGNOP; }
                        else if (   nextChar == '?'
                                 && fsAnlzOptions.GetMode() & ( FSM_HW_HC16 | FSM_HW_851 | FSM_HW_XA | FSM_HW_HC5_11_12 )
                                ) NEXT;
                        goto FoundSTAR;

					case '%':
                  NEXT; if ( nextChar == '=' ) { NEXT; goto FoundASSIGNOP; }
                        else if (   nextChar == '?'
                                 && fsAnlzOptions.GetMode() & ( FSM_HW_HC16 | FSM_HW_851 | FSM_HW_XA | FSM_HW_HC5_11_12 )
                                ) NEXT;
                        goto FoundDIVOP;

               case '&':
                  NEXT; if ( nextChar == '=' ) { NEXT; goto FoundASSIGNOP; }
								else if ( nextChar == '&' ) { NEXT; goto FoundANDAND; }
                        else goto FoundAND;

               case '|':
                  NEXT; if ( nextChar == '=' ) { NEXT; goto FoundASSIGNOP; }
                        else if ( nextChar == '|' ) { NEXT; goto FoundOROR; }
                        else goto FoundOR;

               case '^':
                  NEXT; if ( nextChar == '=' ) { NEXT; goto FoundASSIGNOP; }
                        else goto FoundXOR;


               case '<':
                  NEXT; if ( nextChar == '<' ) { NEXT; if ( nextChar == '=' ) { NEXT; goto FoundASSIGNOP; } else goto FoundSHIFTOP; }
								else if ( nextChar == '=' ) NEXT;
                        goto FoundRELOP;

               case '>':
                  NEXT; if ( nextChar == '>' ) { NEXT; if ( nextChar == '=' ) { NEXT; goto FoundASSIGNOP; } else goto FoundSHIFTOP; }
                        else if ( nextChar == '=' ) NEXT;
                        goto FoundRELOP;

               case '!':
						NEXT; if ( nextChar == '=' ) { NEXT; goto FoundEQUOP; }
                        else goto FoundUNOP;

               case EOF: goto FoundEOF; // no NEXT, because we must update
                                        // the file and macro call stacks
                                        // before reading next character

               case '@':
                  NEXT;
                  if ( fsAnlzOptions.GetMode() & FSM_C166 ) goto Found_C166_REGMASK;
                  else if ( fsAnlzOptions.GetMode() & FSM_COSMIC_6805 ) goto Found_COSMIC_6805_AT;
                  else if ( fsAnlzOptions.GetMode() & FSM_HI_TECH_6805 ) goto Found_AT;
                  else goto FoundIllegalChar;

               default: NEXT; goto FoundIllegalChar;

            }

	MayBeLongLiteral:
            switch ( nextChar )
            {
               case '\'':
						NEXT; goto CharLiteral;

               case '\"':
                  NEXT; goto String;

               default:
                  goto Identifier;
				}

	Identifier:
            switch ( nextChar )
            {
               case LETTER:
               case DIGIT:
                  NEXT; goto Identifier;

               case '$':
                  if ( fsAnlzOptions.GetMode() & FSM_IAR ) { NEXT; goto Identifier; }
                  goto FoundNAME;

               default:
                  goto FoundNAME; // macro & keyword recognition
            }

   Zero:
            switch ( nextChar )
				{
               case DIGIT: goto OctalNumber;

               case 'x':
               case 'X':
                  NEXT; goto HexNumber;

					default:
                  goto AfterNumber;
            }

   OctalNumber:
            switch ( nextChar )
				{
               case DIGIT0TO7: NEXT; goto OctalNumber;

               case '8':
               case '9':
                  NEXT; FSAnzError( E_INV_NUM_CONST ); goto OctalNumber;

               default:
                  goto AfterNumber;
				}

   DecimalNumber:
            switch ( nextChar )
				{
               case DIGIT:
                  NEXT; goto DecimalNumber;

               default:
                  goto AfterNumber;
            }

   HexNumber:
            switch ( nextChar )
            {
               case HEXDIGIT:
                  NEXT; goto HexNumber;

               default:
                  goto OptIntSuffix;
            }

	AfterNumber:
            switch ( nextChar )
            {
               case '.':
                  NEXT;
                  if ( fsAnlzOptions.GetMode() & FSM_IAR ) goto IARDecimals;
                  else goto Decimals;

               case 'e':
					case 'E':
						NEXT; if ( nextChar == '-' || nextChar == '+' ) NEXT; goto Exponent;

               default:
                  goto OptIntSuffix;
            }

   IARDecimals:
   // support for IAR 78000 stupid '.' binary operator to select bitfields in sfrs
   // expressions such as PORTA.0x2 or PORTA .1 or PORTA . 1 or 0x800.1 or 100 . 1 etc.
   // are legal with 78000
   // in IAR mode, we always accept whole new number after a decimal dot, not only decimals
   // there are also some IAR specific grammar productions to support this
         switch ( nextChar )
         {
            case '0':
               NEXT; goto Zero;

            case DIGIT1TO9:
               NEXT; goto DecimalNumber;

            case 'e':
				case 'E':
					NEXT; goto Exponent;

            default:
               goto OptFloatSuffix;
         }

   Decimals:
         switch ( nextChar )
         {
            case DIGIT:
               NEXT; goto Decimals;

            case 'e':
				case 'E':
					NEXT; goto Exponent;

            default:
               goto OptFloatSuffix;
         }

   Exponent:
			switch ( nextChar )
         {
            case DIGIT:
               NEXT; goto Exponent;

				default:                // ne proverava se broj cifara u eksp.
               goto OptFloatSuffix;

         }

   OptFloatSuffix:
         if ( nextChar == 'f' || nextChar == 'F' || nextChar == 'l' || nextChar == 'L' ) NEXT;
         goto FoundFCON;

   OptIntSuffix:
			if ( nextChar == 'u' || nextChar == 'U' ) { NEXT; if ( nextChar == 'L' || nextChar == 'l' ) NEXT; }
         else if ( nextChar == 'l' || nextChar == 'L' ) { NEXT; if ( nextChar == 'u' || nextChar == 'U' ) NEXT; }
         goto FoundICON;

   Dot:
         switch ( nextChar )
         {
				case DIGIT:
               if ( fsAnlzOptions.GetMode() & FSM_IAR ) goto IARDecimals;
               else goto Decimals;

            case '.':
               NEXT; if ( nextChar == '.' ) { NEXT; goto FoundELLIPSIS; }
               goto Dot; // suvisna tacka; parser ce prijaviti gresku

            default:
               goto FoundSTRUCTOP;
         }

   CharLiteral:
         switch ( nextChar )
			{
            case '\'':
               NEXT; goto FoundICON;

            case '\\':
               NEXT; if ( nextChar == '\'' ) NEXT; goto CharLiteral;

            case '\n':
            case EOF:
               FSAnzError( E_UNTERM_CHAR_CONST );
               goto FoundICON;

            default:
               NEXT; goto CharLiteral;
			}

   String:
      switch ( nextChar )
      {
         case '\"':
            NEXT; goto FoundSTRING;

         case '\\':
            NEXT; if ( processBackslashInString && nextChar == '\"') NEXT; goto String;

         case '\n':
         case EOF:
				// there is no NEXT, so we can process \n or EOF
            // in the next call to FSTokenizer::Get()
				OUT( '\"' );
            FSAnzError( E_UNTERM_STR_CONST );
				goto FoundSTRING;

			default:
				NEXT; goto String;
		}

   Semi:
         if ( recognizeAsmComments == FALSE ) goto FoundSEMI;
         switch ( nextChar )
			{
            case '\n':
            case EOF:
               goto Start;

            default:
               NEXT; goto Semi;
			}

	Sharp:
      if ( nextChar == '#' ) { NEXT; goto FoundSHARPSHARP; }
      else goto FoundSHARP;

   Slash:
		switch ( nextChar )
      {
         case '*': NEXT; goto FoundTradComment;

         case '/':
            if ( preProc.cppEnabled == TRUE ) { NEXT; goto FoundCppComment; }
            else goto FoundDIVOP;

         case '?':
            if ( fsAnlzOptions.GetMode() & ( FSM_HW_HC16 | FSM_HW_851 | FSM_HW_XA | FSM_HW_HC5_11_12 ) ) NEXT;
            goto FoundDIVOP;

         default:
            goto FoundDIVOP;
      }

   Colon:
		switch ( nextChar )
      {
         case '>':
            if ( fsAnlzOptions.GetMode() & FSM_MSC ) { NEXT; goto Found_BASE_OPERATOR; }
            else goto FoundCOLON;

         default:
            goto FoundCOLON;
      }

   WhiteSpace:
		switch ( nextChar )
      {
         case WHITE_SPACE:
            NEXT; goto WhiteSpace;

         default:
            goto Start;
      }


//----------------------------------------------------------------------
//                      Action routines
//----------------------------------------------------------------------

   FoundTradComment:
      {
         unsigned nestLevel = 1;
         CLEAR;
         do
         {
            while ( nextChar != '*' && nextChar != '/' && nextChar != EOF ) SKIP;
            if ( nextChar == EOF ) break;
            if ( nextChar == '/' )
            {
               SKIP;
               if ( preProc.nestedComments && nextChar == '*' )
					{
                  nestLevel++;
                  SKIP;
               }
            }
            else /* nextChar == '*' */
            {
               SKIP;
               if ( nextChar == '/' )
               {
                  nestLevel--;
                  SKIP;
               }
				}
         }
         while ( nestLevel != 0 );
         goto Start;
      }

   FoundCppComment:
      CLEAR; LAST;
      while ( nextChar != '\n' && nextChar != EOF ) SKIP;
      goto Start;

   FoundIllegalChar: goto Start;

	FoundNEWLINE: LAST; return NEWLINE;

   FoundSHARP: LAST; return SHARP; // #

   FoundSHARPSHARP: LAST; return SHARPSHARP; // ##

   FoundNAME: LAST; return NAME;   // abc, _123

   FoundSTRING: LAST; return STRING;  // "abc", L"abc"

   FoundICON: LAST; return ( fsAnlzOptions.GetMode() & FSM_IAR ) ? _IAR_CON : ICON;      // 123, 0123UL, 0xabc, 'a', '\0x11',...

   FoundFCON: LAST; return ( fsAnlzOptions.GetMode() & FSM_IAR ) ? _IAR_CON : FCON;      // .123e+10, 123.f , 123e5L

   FoundLP: LAST; return LP;  // (

   FoundRP: LAST; return RP;  // )

   FoundLC: LAST; return LC;  // {

   FoundRC: LAST; return RC;  // }

   FoundLB: LAST; return LB;  // [

   FoundRB: LAST; return RB;  // ]

   FoundSTRUCTOP: LAST; return STRUCTOP; // ., ->

   FoundINCOP: LAST; return INCOP;       // ++, --

   FoundUNOP: LAST; return UNOP;         // !, ~

   FoundSTAR: LAST; return STAR;         // *

   FoundDIVOP: LAST; return DIVOP;       // %, /

   FoundPLUS: LAST; return PLUS;         // +

   FoundMINUS: LAST; return MINUS;       // -

   FoundSHIFTOP: LAST; return SHIFTOP;   // <<, >>

   FoundRELOP: LAST; return RELOP;       // <=, >=, <, >

   FoundEQUOP: LAST; return EQUOP;       // ==, !=

   FoundASSIGNOP: LAST; return ASSIGNOP;  // +=, -=, *=, /=, %=, &=, |=, ^=, <<=, >>=

   FoundEQUAL: LAST; return EQUAL;        // =

   FoundAND: LAST; return AND;           // &

   FoundXOR: LAST; return XOR;           // ^

   FoundOR: LAST; return OR;             // |

   FoundANDAND: LAST; return ANDAND;     // &&

   FoundOROR: LAST; return OROR;         // ||

   FoundQUEST: LAST; return QUEST;       // ?

   Found_BASE_OPERATOR: LAST; return _BASE_OPERATOR;       // :>

   FoundCOLON: LAST; return COLON;       // :

   FoundCOMMA: LAST; return COMMA;       // ,

   FoundSEMI: LAST; return SEMI;         // ;

   FoundELLIPSIS: LAST; return ELLIPSIS; // ...

   Found_C166_REGMASK: LAST; return _C166_REGMASK; // @

   Found_COSMIC_6805_AT: LAST; return _COSMIC_6805_AT; // @

   Found_AT: LAST; return _AT; // @

   FoundEOF: return EOF;

}
