#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <fstream.h>
#if defined(PPDEBUG) || defined(YYDEBUG)
   #include <stdio.h>
#endif

#ifdef DEBUG
   #ifdef BRANKO
      #include "debug.h"
   #endif
#endif

#include "fsbool.h"
#include "fsanzopt.h"
#include "fserror.h"
#include "fsifexpr.h"
#include "fspproc.h"
#include "fsstacks.h"
#include "fstoken.h"
//#include "fssymtab.h" // fspars_t.h needs it
#include "fspars_t.h"
#include "fsprefix.h"

static Boolean tokenIsSaved;
static Boolean expandMacroCalls;
static int token;

static int yyparse();


static void yyerror( char *errStr )
{
   if ( strcmp( errStr, "syntax error" ) ) FSAnzError( E_INV_EXPR );
   else
   {
      FSAnzError( E_PARSER_STACK_OVERFLOW );
      throw 1;
   }
}



//##ModelId=419A04D60173
typedef struct
{
   long value;
   int unsignedp;
} FSPPStackType;

#define FSPPYYSTYPE FSPPStackType
#define FSPP_INT 257
#define FSPP_CHAR 258
#define FSPP_NAME 259
#define FSPP_ERROR 260
#define FSPP_LP 261
#define FSPP_RP 262
#define FSPP_EQUAL 263
#define FSPP_SIZEOF 264
#define FSPP_ARGT 265
#define FSPP_DEFINED 266
#define FSPP_QUEST 267
#define FSPP_COLON 268
#define FSPP_COMMA 269
#define FSPP_OROR 270
#define FSPP_ANDAND 271
#define FSPP_OR 272
#define FSPP_XOR 273
#define FSPP_AND 274
#define FSPP_EQUALEQUAL 275
#define FSPP_NOTEQUAL 276
#define FSPP_LT 277
#define FSPP_GT 278
#define FSPP_LEQ 279
#define FSPP_GEQ 280
#define FSPP_LSH 281
#define FSPP_RSH 282
#define FSPP_PLUS 283
#define FSPP_MINUS 284
#define FSPP_STAR 285
#define FSPP_DIV 286
#define FSPP_MOD 287
#define FSPP_UNARY 288
#define FSPP_RELNOT 289
#define FSPP_LOGNOT 290
#ifndef YYSTYPE
#define YYSTYPE int
#endif
YYSTYPE yylval, yyval;
#define YYERRCODE 256


/* Take care of parsing a number (anything that starts with a digit).
   Set yylval and return the token type;
   len is the number of characters in it.  */

static int parse_number( char *p, int len )
{
   int c;
   unsigned long n = 0;
   int base = 10;
   int digit, largest_digit = 0;
   int spec_long = 0;

   yylval.unsignedp = 0;

   if ( len >= 3 && !strnicmp( p, "0X", 2 ) )
   {
      p += 2;
      base = 16;
      len -= 2;
   }
   else if ( *p == '0' )
      base = 8;

   for ( ; len > 0; len-- )
   {
      c = *p++;

      if ( isdigit( c ) )
         digit = c - '0';
      else if ( base == 16 && isxdigit( c ) )
         digit = toupper( c ) - 'A' + 10;
      else
      {
         /* 'l' means long, and 'u' means unsigned.  */
         while ( 1 )
         {
            if ( c == 'l' || c == 'L' )
            {
               if ( spec_long )
                  FSAnzError( E_INV_NUM_CONST );
               spec_long = 1;
            }
            else if ( c == 'u' || c == 'U' )
            {
               if ( yylval.unsignedp )
                  FSAnzError( E_INV_NUM_CONST );
               yylval.unsignedp = 1;
            }
            else
               break;

            if ( --len == 0 )
               break;
            c = *p++;
         }
         /* Don't look for any more digits after the suffixes.  */
         break;
      }
      if ( largest_digit < digit )
         largest_digit = digit;
      n = n * base + digit;
   }

   if ( len != 0 || base <= largest_digit )
   {
      FSAnzError( E_INV_NUM_CONST );
      return FSPP_ERROR;
   }

   /* If too big to be signed, consider it unsigned.  */
   if ( ( long ) n < 0 && !yylval.unsignedp )
   {
      yylval.unsignedp = 1;
   }
   yylval.value = n;
   return FSPP_INT;
}

static int parse_escape( char **string_ptr )
{
   register int c = *( *string_ptr )++;
   switch ( c )
   {
      case 'n':
         return '\n';
      case 't':
         return '\t';
      case 'r':
         return '\r';
      case 'a':
         return '\a';
      case 'b':
         return '\b';
      case 'f':
         return '\f';
      case 'v':
         return '\v';
      case '\n':
         return -2;
      case '\\':
      case '\'':
      case '"':
      case '?':
         return c;
      case 0:
         ( *string_ptr )--;
         return 0;

      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
         {
            register int i = c - '0';
            register int count = 0;
            while ( ++count < 3 )
            {
               c = *( *string_ptr )++;
               if ( c >= '0' && c <= '7' )
                  i = ( i << 3 ) + c - '0';
               else
               {
                  ( *string_ptr )--;
                  break;
               }
            }
            if ( i > 255 )
            {
               FSAnzError( W_INV_ESCAPE );
               i &= 255;
            }
            return i;
         }
      case 'x':
         {
            register unsigned i = 0, overflow = 0, digits_found = 0, digit;
            for ( ;; )
            {
               c = *( *string_ptr )++;
               if ( isdigit( c ) )
                  digit = c - '0';
               else if ( isxdigit( c ) )
                  digit = toupper( c ) - 'A' + 10;
               else
               {
                  ( *string_ptr )--;
                  break;
               }
               i = ( i << 4 ) + digit;
               digits_found = 1;
               overflow |= ( i > 255 );
            }
            if ( !digits_found || overflow )
            {
               FSAnzError( W_INV_ESCAPE );
               i &= 255;
            }
            return i;
         }
      default:
         //if ( !DialogAnzOptData.PPTraditional )
            FSAnzError( W_INV_ESCAPE );
         return c;
   }
}

// The lexical analyzer for preprocessor if-expressions parser


static int yylex()
{
   while ( 1 )
   {
      if ( tokenIsSaved == FALSE ) token = preProc->tokenizer.Get();
      switch( token )
      {
         case NAME:
            {
					FSMacro *pMacro;
               // first check if operator defined
					// its argument should not be macro expanded
					if ( strcmp( preProc->outBuffer.GetTxt(), "defined" ) == 0 )
					{
						return FSPP_DEFINED;
					}
					// not operator defined; macro expand if necessary
					else if (
									( yylval.value = ( pMacro = preProc->macroTable.LookUp( preProc->outBuffer.GetTxt() ) ) != 0 ) != 0
								&& expandMacroCalls
								&& preProc->macroCallStack.ExpandingNow( pMacro ) == FALSE
							  )
					{
						// there is a macro with that name & it is not recursive call
						if ( pMacro->Args().Parenth() == FALSE )  // macro need no parenthesis
						{
							preProc->macroCallStack.Push( *pMacro );
							break;
						}
						else if ( ( token = preProc->tokenizer.Get() ) == LP )// macro with parenthesis & there are parenth. following a name
						{
							preProc->macroCallStack.Push( *pMacro );
							break;
						}
						else // not a macro call, must remember lookahead
						{
							tokenIsSaved = TRUE;
							return FSPP_NAME;
						}
					}
					// token is not a macro name
					return FSPP_NAME;
            }

         case ICON:
            {
               char *constTxt = preProc->outBuffer.GetTxt();
               if ( *constTxt == 'L' )
               {
                  constTxt++;
               }
               if ( *constTxt != '\'' )
               {
                  // parse_number sets yylval and returns token code for the parser
                  return parse_number( preProc->outBuffer.GetTxt(), preProc->outBuffer.GetLen() );
               }
               else if ( *++constTxt == '\\' ) // skipping '
               {
                  yylval.value = parse_escape( &constTxt );
                  yylval.unsignedp = 0;
               }
               else if ( *constTxt != '\'' && *constTxt != '\0' )
               {
                  yylval.value = *constTxt;
                  yylval.unsignedp = 0;
               }
               else
               {
                  FSAnzError( E_INV_NUM_CONST );
                  yylval.value = 0;
                  yylval.unsignedp = 0;
               }
               return FSPP_INT;
            }

         case UNOP:
            return ( *( preProc->outBuffer.GetTxt() ) == '!' )
                    ? FSPP_RELNOT
                    : FSPP_LOGNOT ;

         case LP:
            return FSPP_LP;

         case RP:
            return FSPP_RP;

         case PLUS:
            return FSPP_PLUS;

         case MINUS:
            return FSPP_MINUS;

         case STAR:
            return FSPP_STAR;

         case SHIFTOP:
            return ( *( preProc->outBuffer.GetTxt() ) == '<' )
                      ? FSPP_LSH
                      : FSPP_RSH;

         case DIVOP:
            return ( *( preProc->outBuffer.GetTxt() ) == '%' )
                    ? FSPP_MOD
                    : FSPP_DIV;

         case RELOP:
            if ( preProc->outBuffer.GetLen() > 1 )
            {
            return ( *( preProc->outBuffer.GetTxt() ) == '>' )
                    ? FSPP_GEQ
                    : FSPP_LEQ;
            }
            return ( *( preProc->outBuffer.GetTxt() ) == '>' )
                    ? FSPP_GT
                    : FSPP_LT;

         case EQUOP:
            return ( *( preProc->outBuffer.GetTxt() ) == '!' )
                    ? FSPP_NOTEQUAL
                    : FSPP_EQUAL;

         case AND:
            return FSPP_AND;

         case XOR:
            return FSPP_XOR;

         case OR:
            return FSPP_OR;

         case ANDAND:
            return FSPP_ANDAND;

         case OROR:
            return FSPP_OROR;

         case COLON:
            return FSPP_COLON;

         case COMMA:
            return FSPP_COMMA;

         case EOF:
            if ( preProc->macroCallStack.IsNotEmpty() )
            {
               preProc->macroCallStack.Pop();
            }
            return EOF;

         case NEWLINE:
            if ( preProc->macroCallStack.IsNotEmpty() )
            {
               break;
            }
            tokenIsSaved = TRUE;
            return EOF;

         default:
            return FSPP_ERROR;
      }
   }
}

static long left_shift( YYSTYPE * a, unsigned long b )
{
   if ( a->unsignedp )
      return ( unsigned long ) a->value << b;
   else
      return a->value << b;
}

static long right_shift( YYSTYPE * a, unsigned long b )
{
   if ( a->unsignedp )
      return ( unsigned long ) a->value >> b;
   else
      return a->value >> b;
}

Boolean ParseIfExpr( int &lastTokenRead )
{
   ::tokenIsSaved = FALSE;
   ::expandMacroCalls = TRUE;
   int result = 1;
   try
   {
      result = yyparse();
   }
   catch(...) {}
   // scan to end of line in case there was an error
   while ( yylex() != EOF )
      ;
   lastTokenRead = ::token;
   return ( Boolean ) !result;
}

#if 0
/* Main program for testing purposes.  */
int main()
{
   int n, c;
   char buf[1024];

   for ( ;; )
   {
      printf( "enter expression: " );
      n = 0;
      while ( ( buf[n] = getchar() ) != '\n' && buf[n] != EOF )
         n++;
      if ( buf[n] == EOF )
         break;
      buf[n] = '\0';
      printf( "parser returned %ld\n", parse_c_expression( buf ) );
   }

   return 0;
}

#endif

const int yyexca[] = {
  -1, 1,
  0, -1,
  -2, 0,
  0,
};

#define YYNPROD 72
#define YYLAST 377

const int yyact[] = {
      95,      92,      98,     102,     105,     110,      85,      16,
      17,      18,      87,      83,      80,     100,      91,      89,
      88,      77,      93,      99,      84,      86,      96,      94,
      97,     101,      79,      81,      78,      82,      76,     108,
      75,      90,      95,      92,      98,     102,     105,     106,
      85,      16,      17,      68,      87,      83,      80,     100,
      91,      89,      88,      77,      93,      99,      84,      86,
      96,      94,      97,     101,      79,      81,      78,      82,
      76,      46,      75,      90,      95,      92,      98,     102,
     105,      74,      85,      16,      17,      45,      87,      83,
      80,     100,      91,      89,      88,      77,      93,      99,
      84,      86,      96,      94,      97,     101,      79,      81,
      78,      82,      76,      72,      75,      90,      37,      71,
      44,      36,      35,      34,      33,      32,      26,      27,
      30,      31,      28,      29,      24,      25,      22,      23,
      19,      20,      21,      37,      73,      43,      36,      35,
      34,      33,      32,      26,      27,      30,      31,      28,
      29,      24,      25,      22,      23,      19,      20,      21,
      35,      34,      33,      32,      26,      27,      30,      31,
      28,      29,      24,      25,      22,      23,      19,      20,
      21,      34,      33,      32,      26,      27,      30,      31,
      28,      29,      24,      25,      22,      23,      19,      20,
      21,      33,      32,      26,      27,      30,      31,      28,
      29,      24,      25,      22,      23,      19,      20,      21,
       9,      10,      11,      14,       8,      13,      12,      16,
      17,      15,      32,      26,      27,      30,      31,      28,
      29,      24,      25,      22,      23,      19,      20,      21,
      69,       1,       6,       4,       0,      19,      20,      21,
       5,       7,      26,      27,      30,      31,      28,      29,
      24,      25,      22,      23,      19,      20,      21,      30,
      31,      28,      29,      24,      25,      22,      23,      19,
      20,      21,      24,      25,      22,      23,      19,      20,
      21,      22,      23,      19,      20,      21,      67,      70,
       2,     104,     104,     103,     103,      18,       0,       0,
       0,      42,       0,       0,       0,       0,       0,       0,
       0,       0,       0,       0,       0,       0,       0,       0,
       0,       0,       0,       0,       0,       0,       0,       0,
       0,       0,       0,       0,       0,       0,       0,       0,
       3,     104,       0,     103,       0,      38,      39,      40,
      41,       0,       0,       0,       0,       0,       0,       0,
       0,       0,     109,      47,      48,      49,      50,      51,
      52,      53,      54,      55,      56,      57,      58,      59,
      60,      61,      62,      63,      64,      65,      66,       0,
       0,       0,       0,       0,       0,       0,       0,       0,
       0,       0,       0,       0,       0,       0,       0,       0,
       0,       0,       0,       0,       0,       0,       0,       0,
       0,       0,       0,       0,       0,       0,       0,       0,
     107,
};

const int yypact[] = {
     -65,   -4096,    -260,    -144,     -65,     -65,     -65,     -65,
     -65,   -4096,   -4096,   -4096,   -4096,   -4096,   -4096,    -152,
    -184,    -196,     -65,     -65,     -65,     -65,     -65,     -65,
     -65,     -65,     -65,     -65,     -65,     -65,     -65,     -65,
     -65,     -65,     -65,     -65,     -65,     -65,   -4096,   -4096,
   -4096,   -4096,       0,    -218,   -4096,   -4096,   -4096,    -144,
   -4096,   -4096,   -4096,     -64,     -64,     -26,     -26,     -38,
     -38,     -31,     -31,     -31,     -31,     -49,     -72,     -96,
    -111,    -127,    -165,   -4096,    -160,    -189,    -223,     -65,
    -231,   -4096,   -4096,   -4096,   -4096,   -4096,   -4096,   -4096,
   -4096,   -4096,   -4096,   -4096,   -4096,   -4096,   -4096,   -4096,
   -4096,   -4096,   -4096,   -4096,   -4096,   -4096,   -4096,   -4096,
   -4096,   -4096,   -4096,   -4096,   -4096,   -4096,   -4096,   -4096,
   -4096,   -4096,   -4096,    -144,   -4096,    -257,   -4096,
};

const int yypgo[] = {
       0,     217,     264,     304,     198,     197,     195,     125,
     124,     216,
};

const int yyr1[] = {
       0,       1,       2,       2,       3,       3,       3,       3,
       3,       3,       3,       3,       3,       3,       3,       3,
       3,       3,       3,       3,       3,       3,       3,       3,
       3,       3,       3,       3,       3,       3,       3,       3,
       3,       3,       7,       4,       4,       8,       8,       8,
       8,       8,       8,       8,       8,       8,       8,       8,
       8,       8,       8,       8,       8,       8,       8,       8,
       8,       8,       8,       8,       8,       8,       8,       8,
       8,       8,       8,       8,       9,       9,       5,       6,
};

const int yyr2[] = {
       0,       1,       1,       3,       2,       2,       2,       2,
       3,       3,       3,       3,       3,       3,       3,       3,
       3,       3,       3,       3,       3,       3,       3,       3,
       3,       3,       3,       5,       1,       1,       1,       1,
       1,       1,       0,       5,       2,       1,       1,       1,
       1,       1,       1,       1,       1,       1,       1,       1,
       1,       1,       1,       1,       1,       1,       1,       1,
       1,       1,       1,       1,       1,       1,       1,       1,
       1,       1,       1,       3,       0,       2,       4,       4,
};

const int yychk[] = {
   -4096,      -1,      -2,      -3,     284,     289,     283,     290,
     261,     257,     258,     259,      -4,      -5,      -6,     266,
     264,     265,     269,     285,     286,     287,     283,     284,
     281,     282,     275,     276,     279,     280,     277,     278,
     274,     273,     272,     271,     270,     267,      -3,      -3,
      -3,      -3,      -2,      -7,     256,     261,     261,      -3,
      -3,      -3,      -3,      -3,      -3,      -3,      -3,      -3,
      -3,      -3,      -3,      -3,      -3,      -3,      -3,      -3,
      -3,      -3,      -3,     262,     261,      -9,      -9,     268,
     259,      -8,     262,     289,     287,     274,     285,     283,
     269,     284,     286,     268,     277,     263,     278,     267,
     273,     272,     290,     271,     258,     275,     280,     257,
     279,     281,     259,     276,     270,     282,     260,      -5,
      -6,     261,     262,      -3,     262,      -9,     262,
};

const int yydef[] = {
       0,      -2,       1,       2,       0,       0,       0,       0,
       0,      28,      29,      30,      31,      32,      33,      34,
       0,       0,       0,       0,       0,       0,       0,       0,
       0,       0,       0,       0,       0,       0,       0,       0,
       0,       0,       0,       0,       0,       0,       4,       5,
       6,       7,       0,       0,      36,      68,      68,       3,
       9,      10,      11,      12,      13,      14,      15,      16,
      17,      18,      19,      20,      21,      22,      23,      24,
      25,      26,       0,       8,       0,       0,       0,       0,
       0,      69,      70,      37,      38,      39,      40,      41,
      42,      43,      44,      45,      46,      47,      48,      49,
      50,      51,      52,      53,      54,      55,      56,      57,
      58,      59,      60,      61,      62,      63,      64,      65,
      66,      68,      71,      27,      35,       0,      67,
};

/*****************************************************************/
/* PCYACC LALR parser driver routine -- a table driven procedure */
/* for recognizing sentences of a language defined by the        */
/* grammar that PCYACC analyzes. An LALR parsing table is then   */
/* constructed for the grammar and the skeletal parser uses the  */
/* table when performing syntactical analysis on input source    */
/* programs. The actions associated with grammar rules are       */
/* inserted into a switch statement for execution.               */
/*****************************************************************/


#ifndef YYMAXDEPTH
#define YYMAXDEPTH 200
#endif
#ifndef YYREDMAX
#define YYREDMAX 1000
#endif
#define PCYYFLAG -4096
#define WAS0ERR 0
#define WAS1ERR 1
#define WAS2ERR 2
#define WAS3ERR 3
#define yyclearin pcyytoken = -1
#define yyerrok   pcyyerrfl = 0
YYSTYPE yyv[YYMAXDEPTH];     /* value stack */
int pcyyerrct = 0;           /* error count */
int pcyyerrfl = 0;           /* error flag */
int redseq[YYREDMAX];
int redcnt = 0;
int pcyytoken = -1;          /* input token */


yyparse()
{
#ifdef YYSHORT
  const short *yyxi;
#else
  const int *yyxi;
#endif
#ifdef YYASTFLAG
    int ti; int tj;
#endif

#ifdef YYDEBUG
int tmptoken;
#endif
  int statestack[YYMAXDEPTH]; /* state stack */
  int      j, m;              /* working index */
  YYSTYPE *yypvt;
  int      tmpstate, *yyps, n;
  YYSTYPE *yypv;


  tmpstate = 0;
  pcyytoken = -1;


#ifdef YYDEBUG
  tmptoken = -1;
#endif


  pcyyerrct = 0;
  pcyyerrfl = 0;
  yyps = &statestack[-1];
  yypv = &yyv[-1];


  enstack:    /* push stack */
#ifdef YYDEBUG
    printf("at state %d, next token %d\n", tmpstate, tmptoken);
#endif
    if (++yyps - &statestack[YYMAXDEPTH-1] > 0) {
      yyerror("pcyacc internal stack overflow");
      return(1);
    }
    *yyps = tmpstate;
    ++yypv;
    *yypv = yyval;


  newstate:
    n = yypact[tmpstate];
    if (n <= PCYYFLAG) goto defaultact; /*  a simple state */


    if (pcyytoken < 0) if ((pcyytoken=yylex()) < 0) pcyytoken = 0;
    if ((n += pcyytoken) < 0 || n >= YYLAST) goto defaultact;


    if (yychk[n=yyact[n]] == pcyytoken) { /* a shift */
#ifdef YYDEBUG
      tmptoken  = pcyytoken;
#endif
      pcyytoken = -1;
      yyval = yylval;
      tmpstate = n;
      if (pcyyerrfl > 0) --pcyyerrfl;
      goto enstack;
    }


  defaultact:


    if ((n=yydef[tmpstate]) == -2) {
      if (pcyytoken < 0) if ((pcyytoken=yylex())<0) pcyytoken = 0;
      for (yyxi=yyexca; (*yyxi!= (-1)) || (yyxi[1]!=tmpstate); yyxi += 2);
      while (*(yyxi+=2) >= 0) if (*yyxi == pcyytoken) break;
      if ((n=yyxi[1]) < 0) { /* an accept action */
#ifdef YYASTFLAG
          yytfilep = fopen(yytfilen, "w");
          if (yytfilep == NULL) {
            fprintf(stderr, "Can't open t file: %s\n", yytfilen);
            return(0);          }
          for (ti=redcnt-1; ti>=0; ti--) {
            tj = svdprd[redseq[ti]];
            while (strcmp(svdnams[tj], "$EOP"))
              fprintf(yytfilep, "%s ", svdnams[tj++]);
            fprintf(yytfilep, "\n");
          }
          fclose(yytfilep);
#endif

        return (0);
      }
    }


    if (n == 0) {        /* error situation */
      switch (pcyyerrfl) {
        case WAS0ERR:          /* an error just occurred */
          yyerror("syntax error");
            ++pcyyerrct;
        case WAS1ERR:
        case WAS2ERR:           /* try again */
          pcyyerrfl = 3;
	   /* find a state for a legal shift action */
          while (yyps >= statestack) {
	     n = yypact[*yyps] + YYERRCODE;
	     if (n >= 0 && n < YYLAST && yychk[yyact[n]] == YYERRCODE) {
	       tmpstate = yyact[n];  /* simulate a shift of "error" */
	       goto enstack;
            }
	     n = yypact[*yyps];


	     /* the current yyps has no shift on "error", pop stack */
#ifdef YYDEBUG
            printf("error: pop state %d, recover state %d\n", *yyps, yyps[-1]);
#endif
	     --yyps;
	     --yypv;
	   }


	   yyabort:

#ifdef YYASTFLAG
              yytfilep = fopen(yytfilen, "w");
              if (yytfilep == NULL) {
                fprintf(stderr, "Can't open t file: %s\n", yytfilen);
                return(1);              }
              for (ti=1; ti<redcnt; ti++) {
                tj = svdprd[redseq[ti]];
                while (strcmp(svdnams[tj], "$EOP"))
                  fprintf(yytfilep, "%s ", svdnams[tj++]);
                fprintf(yytfilep, "\n");
              }
              fclose(yytfilep);
#endif

	     return(1);


	 case WAS3ERR:  /* clobber input char */
#ifdef YYDEBUG
          printf("error: discard token %d\n", pcyytoken);
#endif
          if (pcyytoken == 0) goto yyabort; /* quit */
	   pcyytoken = -1;
	   goto newstate;      } /* switch */
    } /* if */


    /* reduction, given a production n */
#ifdef YYDEBUG
    printf("reduce with rule %d\n", n);
#endif
#ifdef YYASTFLAG
    if ( redcnt<YYREDMAX ) redseq[redcnt++] = n;
#endif
    yyps -= yyr2[n];
    yypvt = yypv;
    yypv -= yyr2[n];
    yyval = yypv[1];
    m = n;
    /* find next state from goto table */
    n = yyr1[n];
    j = yypgo[n] + *yyps + 1;
    if (j>=YYLAST || yychk[ tmpstate = yyact[j] ] != -n) tmpstate = yyact[yypgo[n]];
    switch (m) { /* actions associated with grammar rules */
      
      case 1:{ return ( int )( yypvt[-0].value != 0 ); } break;
      case 3:{ yyval = yypvt[-0];} break;
      case 4:{  yyval.value = -yypvt[-0].value;
                 yyval.unsignedp = yypvt[-0].unsignedp;
              } break;
      case 5:{
                  yyval.value = !yypvt[-0].value;
                 yyval.unsignedp = 0;
              } break;
      case 6:{ yyval = yypvt[-0]; } break;
      case 7:{
                 yyval.value = ~yypvt[-0].value;
                 yyval.unsignedp = yypvt[-0].unsignedp;
              } break;
      case 8:{ yyval = yypvt[-1]; } break;
      case 9:{  yyval.unsignedp = yypvt[-2].unsignedp || yypvt[-0].unsignedp;
                 if ( yyval.unsignedp )
                    yyval.value = ( unsigned long ) yypvt[-2].value * yypvt[-0].value;
                 else
                    yyval.value = yypvt[-2].value * yypvt[-0].value;
              } break;
      case 10:{  if ( yypvt[-0].value == 0 )
                 {
                    FSAnzError( E_DIV_ZERO );
                    yypvt[-0].value = 1;
                 }
                 yyval.unsignedp = yypvt[-2].unsignedp || yypvt[-0].unsignedp;
                 if ( yyval.unsignedp )
                    yyval.value = ( unsigned long ) yypvt[-2].value / yypvt[-0].value;
                 else
                    yyval.value = yypvt[-2].value / yypvt[-0].value;
              } break;
      case 11:{  if ( yypvt[-0].value == 0 )
                 {
                    FSAnzError( E_DIV_ZERO );
                    yypvt[-0].value = 1;
                 }
                 yyval.unsignedp = yypvt[-2].unsignedp || yypvt[-0].unsignedp;
                 if ( yyval.unsignedp )
                    yyval.value = ( unsigned long ) yypvt[-2].value % yypvt[-0].value;
                 else
                    yyval.value = yypvt[-2].value % yypvt[-0].value;
              } break;
      case 12:{  yyval.value = yypvt[-2].value + yypvt[-0].value;
                 yyval.unsignedp = yypvt[-2].unsignedp || yypvt[-0].unsignedp;
              } break;
      case 13:{  yyval.value = yypvt[-2].value - yypvt[-0].value;
                 yyval.unsignedp = yypvt[-2].unsignedp || yypvt[-0].unsignedp;
              } break;
      case 14:{  yyval.unsignedp = yypvt[-2].unsignedp;
                 if ( yypvt[-0].value < 0 && !yypvt[-0].unsignedp )
                    yyval.value = right_shift( &yypvt[-2], -yypvt[-0].value );
                 else
                    yyval.value = left_shift( &yypvt[-2], yypvt[-0].value );
              } break;
      case 15:{  yyval.unsignedp = yypvt[-2].unsignedp;
                 if ( yypvt[-0].value < 0 && !yypvt[-0].unsignedp )
                    yyval.value = left_shift( &yypvt[-2], -yypvt[-0].value );
                 else
                    yyval.value = right_shift( &yypvt[-2], yypvt[-0].value );
              } break;
      case 16:{  yyval.value = ( yypvt[-2].value == yypvt[-0].value );
                 yyval.unsignedp = 0;
              } break;
      case 17:{  yyval.value = ( yypvt[-2].value != yypvt[-0].value );
                 yyval.unsignedp = 0;
              } break;
      case 18:{  yyval.unsignedp = 0;
                 if ( yypvt[-2].unsignedp || yypvt[-0].unsignedp )
                    yyval.value = ( unsigned long ) yypvt[-2].value <= yypvt[-0].value;
                 else
                    yyval.value = yypvt[-2].value <= yypvt[-0].value;
              } break;
      case 19:{  yyval.unsignedp = 0;
                 if ( yypvt[-2].unsignedp || yypvt[-0].unsignedp )
                    yyval.value = ( unsigned long ) yypvt[-2].value >= yypvt[-0].value;
                 else
                    yyval.value = yypvt[-2].value >= yypvt[-0].value;
              } break;
      case 20:{  yyval.unsignedp = 0;
                 if ( yypvt[-2].unsignedp || yypvt[-0].unsignedp )
                    yyval.value = ( unsigned long ) yypvt[-2].value < yypvt[-0].value;
                 else
                    yyval.value = yypvt[-2].value < yypvt[-0].value;
              } break;
      case 21:{  yyval.unsignedp = 0;
                 if ( yypvt[-2].unsignedp || yypvt[-0].unsignedp )
                    yyval.value = ( unsigned long ) yypvt[-2].value > yypvt[-0].value;
                 else
                    yyval.value = yypvt[-2].value > yypvt[-0].value;
              } break;
      case 22:{  yyval.value = yypvt[-2].value & yypvt[-0].value;
                 yyval.unsignedp = yypvt[-2].unsignedp || yypvt[-0].unsignedp;
              } break;
      case 23:{  yyval.value = yypvt[-2].value ^ yypvt[-0].value;
                 yyval.unsignedp = yypvt[-2].unsignedp || yypvt[-0].unsignedp;
              } break;
      case 24:{  yyval.value = yypvt[-2].value | yypvt[-0].value;
                 yyval.unsignedp = yypvt[-2].unsignedp || yypvt[-0].unsignedp;
              } break;
      case 25:{  yyval.value = ( yypvt[-2].value && yypvt[-0].value );
                 yyval.unsignedp = 0;
              } break;
      case 26:{  yyval.value = ( yypvt[-2].value || yypvt[-0].value );
                 yyval.unsignedp = 0;
              } break;
      case 27:{  yyval.value = yypvt[-4].value ? yypvt[-2].value : yypvt[-0].value;
                 yyval.unsignedp = yypvt[-2].unsignedp || yypvt[-0].unsignedp;
              } break;
      case 28:{ yyval = yypvt[-0]; } break;
      case 29:{ yyval = yypvt[-0]; } break;
      case 30:{  yyval.value = 0;
                 yyval.unsignedp = 0;
              } break;
      case 32:{  yyval.value = 1;
                 yyval.unsignedp = 0;
              } break;
      case 33:{  yyval.value = 1;
                 yyval.unsignedp = 0;
              } break;
      case 34:{ expandMacroCalls = FALSE; } break;
      case 35:{  expandMacroCalls = TRUE;
                  yyval.value = yypvt[-1].value;
                  yyval.unsignedp = 0;
               } break;
      case 36:{  expandMacroCalls = TRUE;
                  yyval.value = 0;
                  yyval.unsignedp = 0;
                  FSAnzError( E_INV_EXPR );
               } break;
      case 70:{ /*SizeofInPPExpr = 1; */} break;
      case 71:{ /*ArgtInPPExpr = 1;*/ } break;    }
    goto enstack;
}
