//-----------------------------------------------------------------------
//  Copyright (c) by RPS Professional Software
//-----------------------------------------------------------------------
//  Project name: SOKO
//  Module file : fsifexpr.y
//  Compiler    : Borland C++ 4.0
//  Copied from : Adapted from pp_exp.y by B. Marovic
//  Created by  : Dragan Bojic
//-----------------------------------------------------------------------
//  Description :
//
// Parse #if expressions for C preprocessor.
//-----------------------------------------------------------------------
//  Changes:
//  $Log__$
//-----------------------------------------------------------------------

%{
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <fstream.h>
#if defined(PPDEBUG) || defined(YYDEBUG)
   #include <stdio.h>
#endif

#ifdef DEBUG
   #ifdef BRANKO
      #include "debug.h"
   #endif
#endif

#include "fsbool.h"
#include "fsanzopt.h"
#include "fserror.h"
#include "fsifexpr.h"
#include "fspproc.h"
#include "fsstacks.h"
#include "fstoken.h"
#include "fssymtab.h" // fspars_t.h needs it
#include "fspars_t.h"
#include "fsprefix.h"

static Boolean tokenIsSaved;
static Boolean expandMacroCalls;
static int token;

static int yyparse();


static void yyerror( char *errStr )
{
   if ( strcmp( errStr, "syntax error" ) ) FSAnzError( E_INV_EXPR );
   else
   {
      FSAnzError( E_PARSER_STACK_OVERFLOW );
      throw 1;
   }
}


// eliminate 'unreachable code' warnings
#pragma warn -rch

typedef struct
{
   long value;
   int unsignedp;
} FSPPStackType;

#define FSPPYYSTYPE FSPPStackType
%}

%token FSPP_INT
%token FSPP_CHAR
%token FSPP_NAME
%token FSPP_ERROR
%token FSPP_LP
%token FSPP_RP
%token FSPP_EQUAL

%token FSPP_SIZEOF
%token FSPP_ARGT
%token FSPP_DEFINED

%right FSPP_QUEST FSPP_COLON
%left FSPP_COMMA
%left FSPP_OROR
%left FSPP_ANDAND
%left FSPP_OR
%left FSPP_XOR
%left FSPP_AND
%left FSPP_EQUALEQUAL FSPP_NOTEQUAL
%left FSPP_LT FSPP_GT FSPP_LEQ FSPP_GEQ
%left FSPP_LSH FSPP_RSH
%left FSPP_PLUS FSPP_MINUS
%left FSPP_STAR FSPP_DIV FSPP_MOD
%right FSPP_UNARY FSPP_RELNOT FSPP_LOGNOT

%%

start:  exp1 { return ( int )( $1.value != 0 ); }
      ;

/* Expressions, including the comma operator.  */
exp1:   exp 
      | exp1 FSPP_COMMA exp { $$ = $3;}
      ;

/* Expressions, not including the comma operator.  */
exp:    FSPP_MINUS exp %prec FSPP_UNARY
        {  $$.value = -$2.value;
           $$.unsignedp = $2.unsignedp;
        }
      | FSPP_RELNOT exp %prec FSPP_UNARY
        {
            $$.value = !$2.value;
           $$.unsignedp = 0;
        }
      | FSPP_PLUS exp %prec FSPP_UNARY { $$ = $2; }
      | FSPP_LOGNOT exp %prec FSPP_UNARY
        {
           $$.value = ~$2.value;
           $$.unsignedp = $2.unsignedp;
        }
      | FSPP_LP exp1 FSPP_RP { $$ = $2; }

/* Binary operators in order of decreasing precedence.  */
      |  exp FSPP_STAR exp
        {  $$.unsignedp = $1.unsignedp || $3.unsignedp;
           if ( $$.unsignedp )
              $$.value = ( unsigned long ) $1.value * $3.value;
           else
              $$.value = $1.value * $3.value;
        }
      | exp FSPP_DIV exp
        {  if ( $3.value == 0 )
           {
              FSAnzError( E_DIV_ZERO );
              $3.value = 1;
           }
           $$.unsignedp = $1.unsignedp || $3.unsignedp;
           if ( $$.unsignedp )
              $$.value = ( unsigned long ) $1.value / $3.value;
           else
              $$.value = $1.value / $3.value;
        }
      | exp FSPP_MOD exp
        {  if ( $3.value == 0 )
           {
              FSAnzError( E_DIV_ZERO );
              $3.value = 1;
           }
           $$.unsignedp = $1.unsignedp || $3.unsignedp;
           if ( $$.unsignedp )
              $$.value = ( unsigned long ) $1.value % $3.value;
           else
              $$.value = $1.value % $3.value;
        }
      | exp FSPP_PLUS exp
        {  $$.value = $1.value + $3.value;
           $$.unsignedp = $1.unsignedp || $3.unsignedp;
        }
      | exp FSPP_MINUS exp
        {  $$.value = $1.value - $3.value;
           $$.unsignedp = $1.unsignedp || $3.unsignedp;
        }
      | exp FSPP_LSH exp
        {  $$.unsignedp = $1.unsignedp;
           if ( $3.value < 0 && !$3.unsignedp )
              $$.value = right_shift( &$1, -$3.value );
           else
              $$.value = left_shift( &$1, $3.value );
        }
      | exp FSPP_RSH exp
        {  $$.unsignedp = $1.unsignedp;
           if ( $3.value < 0 && !$3.unsignedp )
              $$.value = left_shift( &$1, -$3.value );
           else
              $$.value = right_shift( &$1, $3.value );
        }
      | exp FSPP_EQUALEQUAL exp
        {  $$.value = ( $1.value == $3.value );
           $$.unsignedp = 0;
        }
      | exp FSPP_NOTEQUAL exp
        {  $$.value = ( $1.value != $3.value );
           $$.unsignedp = 0;
        }
      | exp FSPP_LEQ exp
        {  $$.unsignedp = 0;
           if ( $1.unsignedp || $3.unsignedp )
              $$.value = ( unsigned long ) $1.value <= $3.value;
           else
              $$.value = $1.value <= $3.value;
        }
      | exp FSPP_GEQ exp
        {  $$.unsignedp = 0;
           if ( $1.unsignedp || $3.unsignedp )
              $$.value = ( unsigned long ) $1.value >= $3.value;
           else
              $$.value = $1.value >= $3.value;
        }
      | exp FSPP_LT exp
        {  $$.unsignedp = 0;
           if ( $1.unsignedp || $3.unsignedp )
              $$.value = ( unsigned long ) $1.value < $3.value;
           else
              $$.value = $1.value < $3.value;
        }
      | exp FSPP_GT exp
        {  $$.unsignedp = 0;
           if ( $1.unsignedp || $3.unsignedp )
              $$.value = ( unsigned long ) $1.value > $3.value;
           else
              $$.value = $1.value > $3.value;
        }
      | exp FSPP_AND exp
        {  $$.value = $1.value & $3.value;
           $$.unsignedp = $1.unsignedp || $3.unsignedp;
        }
      | exp FSPP_XOR exp
        {  $$.value = $1.value ^ $3.value;
           $$.unsignedp = $1.unsignedp || $3.unsignedp;
        }
      | exp FSPP_OR exp
        {  $$.value = $1.value | $3.value;
           $$.unsignedp = $1.unsignedp || $3.unsignedp;
        }
      | exp FSPP_ANDAND exp
        {  $$.value = ( $1.value && $3.value );
           $$.unsignedp = 0;
        }
      | exp FSPP_OROR exp
        {  $$.value = ( $1.value || $3.value );
           $$.unsignedp = 0;
        }
      | exp FSPP_QUEST exp FSPP_COLON exp
        {  $$.value = $1.value ? $3.value : $5.value;
           $$.unsignedp = $3.unsignedp || $5.unsignedp;
        }
      | FSPP_INT { $$ = $1; }
      | FSPP_CHAR { $$ = $1; }
      | FSPP_NAME
        {  $$.value = 0;
           $$.unsignedp = 0;
        }
      | defined_expr
      | so_expr
        {  $$.value = 1;
           $$.unsignedp = 0;
        }
      | _argt_expr
        {  $$.value = 1;
           $$.unsignedp = 0;
        }
        ;

defined_expr:
        FSPP_DEFINED { expandMacroCalls = FALSE; } FSPP_LP FSPP_NAME FSPP_RP
         {  expandMacroCalls = TRUE;
            $$.value = $4.value;
            $$.unsignedp = 0;
         }
      | FSPP_DEFINED error
         {  expandMacroCalls = TRUE;
            $$.value = 0;
            $$.unsignedp = 0;
            FSAnzError( E_INV_EXPR );
         }
      ;

so_elem:
        FSPP_RELNOT
      | FSPP_MOD
      | FSPP_AND
      | FSPP_STAR
      | FSPP_PLUS
      | FSPP_COMMA
      | FSPP_MINUS
      | FSPP_DIV
      | FSPP_COLON
      | FSPP_LT
      | FSPP_EQUAL
      | FSPP_GT
      | FSPP_QUEST
      | FSPP_XOR
      | FSPP_OR
      | FSPP_LOGNOT
      | FSPP_ANDAND
      | FSPP_CHAR
      | FSPP_EQUALEQUAL
      | FSPP_GEQ
      | FSPP_INT
      | FSPP_LEQ
      | FSPP_LSH
      | FSPP_NAME
      | FSPP_NOTEQUAL
      | FSPP_OROR
      | FSPP_RSH
      | FSPP_ERROR
      | so_expr
      | _argt_expr
      | FSPP_LP so_elem_list FSPP_RP
      ;

so_elem_list:
        /* Empty */
      | so_elem_list so_elem
      ;

so_expr:
      FSPP_SIZEOF FSPP_LP so_elem_list FSPP_RP { /*SizeofInPPExpr = 1; */}
      ;

_argt_expr:
      FSPP_ARGT FSPP_LP so_elem_list FSPP_RP { /*ArgtInPPExpr = 1;*/ }
      ;

%%

/* Take care of parsing a number (anything that starts with a digit).
   Set yylval and return the token type;
   len is the number of characters in it.  */

static int parse_number( char *p, int len )
{
   int c;
   unsigned long n = 0;
   int base = 10;
   int digit, largest_digit = 0;
   int spec_long = 0;

   yylval.unsignedp = 0;

   if ( len >= 3 && !strnicmp( p, "0X", 2 ) )
   {
      p += 2;
      base = 16;
      len -= 2;
   }
   else if ( *p == '0' )
      base = 8;

   for ( ; len > 0; len-- )
   {
      c = *p++;

      if ( isdigit( c ) )
         digit = c - '0';
      else if ( base == 16 && isxdigit( c ) )
         digit = toupper( c ) - 'A' + 10;
      else
      {
         /* 'l' means long, and 'u' means unsigned.  */
         while ( 1 )
         {
            if ( c == 'l' || c == 'L' )
            {
               if ( spec_long )
                  FSAnzError( E_INV_NUM_CONST );
               spec_long = 1;
            }
            else if ( c == 'u' || c == 'U' )
            {
               if ( yylval.unsignedp )
                  FSAnzError( E_INV_NUM_CONST );
               yylval.unsignedp = 1;
            }
            else
               break;

            if ( --len == 0 )
               break;
            c = *p++;
         }
         /* Don't look for any more digits after the suffixes.  */
         break;
      }
      if ( largest_digit < digit )
         largest_digit = digit;
      n = n * base + digit;
   }

   if ( len != 0 || base <= largest_digit )
   {
      FSAnzError( E_INV_NUM_CONST );
      return FSPP_ERROR;
   }

   /* If too big to be signed, consider it unsigned.  */
   if ( ( long ) n < 0 && !yylval.unsignedp )
   {
      yylval.unsignedp = 1;
   }
   yylval.value = n;
   return FSPP_INT;
}

static int parse_escape( char **string_ptr )
{
   register int c = *( *string_ptr )++;
   switch ( c )
   {
      case 'n':
         return '\n';
      case 't':
         return '\t';
      case 'r':
         return '\r';
      case 'a':
         return '\a';
      case 'b':
         return '\b';
      case 'f':
         return '\f';
      case 'v':
         return '\v';
      case '\n':
         return -2;
      case '\\':
      case '\'':
      case '"':
      case '?':
         return c;
      case 0:
         ( *string_ptr )--;
         return 0;

      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
         {
            register int i = c - '0';
            register int count = 0;
            while ( ++count < 3 )
            {
               c = *( *string_ptr )++;
               if ( c >= '0' && c <= '7' )
                  i = ( i << 3 ) + c - '0';
               else
               {
                  ( *string_ptr )--;
                  break;
               }
            }
            if ( i > 255 )
            {
               FSAnzError( W_INV_ESCAPE );
               i &= 255;
            }
            return i;
         }
      case 'x':
         {
            register unsigned i = 0, overflow = 0, digits_found = 0, digit;
            for ( ;; )
            {
               c = *( *string_ptr )++;
               if ( isdigit( c ) )
                  digit = c - '0';
               else if ( isxdigit( c ) )
                  digit = toupper( c ) - 'A' + 10;
               else
               {
                  ( *string_ptr )--;
                  break;
               }
               i = ( i << 4 ) + digit;
               digits_found = 1;
               overflow |= ( i > 255 );
            }
            if ( !digits_found || overflow )
            {
               FSAnzError( W_INV_ESCAPE );
               i &= 255;
            }
            return i;
         }
      default:
         //if ( !DialogAnzOptData.PPTraditional )
            FSAnzError( W_INV_ESCAPE );
         return c;
   }
}

// The lexical analyzer for preprocessor if-expressions parser


static int yylex()
{
   while ( 1 )
   {
      if ( tokenIsSaved == FALSE ) token = preProc->tokenizer.Get();
      switch( token )
      {
         case NAME:
            {
					FSMacro *pMacro;
               // first check if operator defined
					// its argument should not be macro expanded
					if ( strcmp( preProc->outBuffer.GetTxt(), "defined" ) == 0 )
					{
						return FSPP_DEFINED;
					}
					// not operator defined; macro expand if necessary
					else if (
									( yylval.value = ( pMacro = preProc->macroTable.LookUp( preProc->outBuffer.GetTxt() ) ) != 0 ) != 0
								&& expandMacroCalls
								&& preProc->macroCallStack.ExpandingNow( pMacro ) == FALSE
							  )
					{
						// there is a macro with that name & it is not recursive call
						if ( pMacro->Args().Parenth() == FALSE )  // macro need no parenthesis
						{
							preProc->macroCallStack.Push( *pMacro );
							break;
						}
						else if ( ( token = preProc->tokenizer.Get() ) == LP )// macro with parenthesis & there are parenth. following a name
						{
							preProc->macroCallStack.Push( *pMacro );
							break;
						}
						else // not a macro call, must remember lookahead
						{
							tokenIsSaved = TRUE;
							return FSPP_NAME;
						}
					}
					// token is not a macro name
					return FSPP_NAME;
            }

         case ICON:
            {
               char *constTxt = preProc->outBuffer.GetTxt();
               if ( *constTxt == 'L' )
               {
                  constTxt++;
               }
               if ( *constTxt != '\'' )
               {
                  // parse_number sets yylval and returns token code for the parser
                  return parse_number( preProc->outBuffer.GetTxt(), preProc->outBuffer.GetLen() );
               }
               else if ( *++constTxt == '\\' ) // skipping '
               {
                  yylval.value = parse_escape( &constTxt );
                  yylval.unsignedp = 0;
               }
               else if ( *constTxt != '\'' && *constTxt != '\0' )
               {
                  yylval.value = *constTxt;
                  yylval.unsignedp = 0;
               }
               else
               {
                  FSAnzError( E_INV_NUM_CONST );
                  yylval.value = 0;
                  yylval.unsignedp = 0;
               }
               return FSPP_INT;
            }

         case UNOP:
            return ( *( preProc->outBuffer.GetTxt() ) == '!' )
                    ? FSPP_RELNOT
                    : FSPP_LOGNOT ;

         case LP:
            return FSPP_LP;

         case RP:
            return FSPP_RP;

         case PLUS:
            return FSPP_PLUS;

         case MINUS:
            return FSPP_MINUS;

         case STAR:
            return FSPP_STAR;

         case SHIFTOP:
            return ( *( preProc->outBuffer.GetTxt() ) == '<' )
                      ? FSPP_LSH
                      : FSPP_RSH;

         case DIVOP:
            return ( *( preProc->outBuffer.GetTxt() ) == '%' )
                    ? FSPP_MOD
                    : FSPP_DIV;

         case RELOP:
            if ( preProc->outBuffer.GetLen() > 1 )
            {
            return ( *( preProc->outBuffer.GetTxt() ) == '>' )
                    ? FSPP_GEQ
                    : FSPP_LEQ;
            }
            return ( *( preProc->outBuffer.GetTxt() ) == '>' )
                    ? FSPP_GT
                    : FSPP_LT;

         case EQUOP:
            return ( *( preProc->outBuffer.GetTxt() ) == '!' )
                    ? FSPP_NOTEQUAL
                    : FSPP_EQUAL;

         case AND:
            return FSPP_AND;

         case XOR:
            return FSPP_XOR;

         case OR:
            return FSPP_OR;

         case ANDAND:
            return FSPP_ANDAND;

         case OROR:
            return FSPP_OROR;

         case COLON:
            return FSPP_COLON;

         case COMMA:
            return FSPP_COMMA;

         case EOF:
            if ( preProc->macroCallStack.IsNotEmpty() )
            {
               preProc->macroCallStack.Pop();
            }
            return EOF;

         case NEWLINE:
            if ( preProc->macroCallStack.IsNotEmpty() )
            {
               break;
            }
            tokenIsSaved = TRUE;
            return EOF;

         default:
            return FSPP_ERROR;
      }
   }
}

static long left_shift( YYSTYPE * a, unsigned long b )
{
   if ( a->unsignedp )
      return ( unsigned long ) a->value << b;
   else
      return a->value << b;
}

static long right_shift( YYSTYPE * a, unsigned long b )
{
   if ( a->unsignedp )
      return ( unsigned long ) a->value >> b;
   else
      return a->value >> b;
}

Boolean ParseIfExpr( int &lastTokenRead )
{
   ::tokenIsSaved = FALSE;
   ::expandMacroCalls = TRUE;
   int result = 1;
   try
   {
      result = yyparse();
   }
   catch(...) {}
   // scan to end of line in case there was an error
   while ( yylex() != EOF )
      ;
   lastTokenRead = ::token;
   return ( Boolean ) !result;
}

#if 0
/* Main program for testing purposes.  */
int main()
{
   int n, c;
   char buf[1024];

   for ( ;; )
   {
      printf( "enter expression: " );
      n = 0;
      while ( ( buf[n] = getchar() ) != '\n' && buf[n] != EOF )
         n++;
      if ( buf[n] == EOF )
         break;
      buf[n] = '\0';
      printf( "parser returned %ld\n", parse_c_expression( buf ) );
   }

   return 0;
}

#endif

