//----------------------------------------------------------------------------
//  Description :
//      type representation & symbol table classess implementation
//----------------------------------------------------------------------------
//  Changes:
//  $Log__$
//----------------------------------------------------------------------------

#include <stdlib.h>
#include <string.h>

#ifdef DEBUG
   #include <iostream.h>
#endif

#ifdef DEBUG
   #ifdef BRANKO
      #include "debug.h"
   #endif
#endif

#include "fsbool.h"
#include "fserror.h"
#include "fssymtab.h"
#include "fsanzopt.h"
#include "fslex.h"

#ifdef DEBUG
   #include "fspproc.h" // for the current line
#endif

/**********************************************************************
 * class Type & derived classes implementation                        *
 **********************************************************************/

/*-------------------------- class BaseType --------------------------*/

/* selectors */
#ifdef DEBUG
//##ModelId=4199F86A0081
void FSBaseType::Print() const
{
   //cout<<"  Type:T_BASE  T:"<<theType<<
   //" C:"<<theClass;

   if ( theClass & -FS_AUTO ) cout<<"auto ";
   if ( theClass & -FS_EXTERN ) cout<<"extern ";
   if ( theClass & -FS_REGISTER ) cout<<"register ";
   if ( theClass & -FS_STATIC ) cout<<"static ";
   if ( theClass & -FS_TYPEDEF ) cout<<"typedef ";

   if ( theType & FS_SIGNED ) cout<<"signed ";
   if ( theType & FS_UNSIGNED ) cout<<"unsigned ";
   if ( theType & FS_LONG ) cout<<"long ";
   if ( theType & FS_SHORT ) cout<<"short ";
   if ( theType & FS_INT ) cout<<"int ";
   if ( theType & FS_CHAR ) cout<<"char ";
   if ( theType & FS_FLOAT ) cout<<"float ";
   if ( theType & FS_DOUBLE ) cout<<"double ";
   if ( theType & FS_VOID ) cout<<"void ";

   if ( GetId() == FS_T_ENUM ) cout<<"enum ";
   if ( GetId() == FS_T_STRUCT ) cout<<"struct ";
   if ( GetId() == FS_T_UNION ) cout<<"union ";
}
#endif

//##ModelId=4199F86A0096
void FSBaseType::CheckTypeDef(Boolean &tDefAllowed) const
{
   const FSBaseTypeSpecifiers hasType = (FSBaseTypeSpecifiers)( FS_INT | FS_CHAR | FS_FLOAT | FS_DOUBLE | FS_VOID
   | FS_LONG | FS_SHORT | FS_SIGNED | FS_UNSIGNED );
   if (theType & hasType) tDefAllowed = FALSE;
}

/* mutators */
static void inline MergeTheType(unsigned &theType, unsigned spec)
{
   /* treba ugraditi druge provere */
   if (theType & spec) FSAnzError ( E_TOO_MANY_TYPESPECS );
   theType = theType | spec;
}

static void inline MergeTheClass(unsigned &theClass, unsigned spec)
{
   /* treba ugraditi druge provere */
   if (theClass & spec) FSAnzError ( E_TOO_MANY_STORAGESPECS );
   theClass = theClass | spec;
}

//##ModelId=4199F86A00A0
FSType *FSBaseType::Merge(FSType *dst, FSType *src)
{
   FSBaseType *srcBase = (FSBaseType *) (src->GetBaseType());
   FSBaseType *dstBase = (FSBaseType *) (dst->GetBaseType());
   if (src->GetId() == FS_T_BASE)
   {
      MergeTheType(dstBase->theType,srcBase->theType);
      MergeTheClass(dstBase->theClass,srcBase->theClass);
      delete src;
      return dst;
   }
   else if (dst->GetId() == FS_T_BASE)
      {
         MergeTheType(srcBase->theType,dstBase->theType);
         MergeTheClass(srcBase->theClass,dstBase->theClass);
         delete dst;
         return src;
      }
      else { /* error - eg. enum enum */
         FSAnzError( E_TOO_MANY_TYPESPECS );
         return src;
      }
}

//##ModelId=4199F86A00B2
void FSBaseType::AddSpecifier(FSBaseTypeSpecifiers spec)
{
   if (spec > 0 ) MergeTheType(theType, spec);
   else MergeTheClass(theClass, -spec);
}

//##ModelId=4199F86A00BC
void FSBaseType::RemoveSpecifier(FSBaseTypeSpecifiers spec)
{
   if (spec > 0 ) theType &= ~spec;
   else theClass &= ~(-spec);
}

/* ------------------------ class DeclarType ------------------------*/

FSSymbol *FSAppendType( FSType *specType, FSSymbol *sym,FSCopySpecFlag cpySpec)
{
   /* povratni tip nastaje dodavanjem kopije specType, kao osnovnog tipa
      na declType */
   /* ako je specType bez BaseType, onda poslednji tip ukazuje sam na sebe
      (theBase) */
   FSType *ptrCopy = (cpySpec == FS_SPEC_COPY) ? specType->Copy() : specType;
   FSType *declType = sym->GetType();
   FSDeclarType *ptrLast;
   if (declType == 0)
   {
      sym->SetType( ptrCopy );
   }
   else
   {
      ptrLast = (FSDeclarType *) ( declType->GetBaseType() );
      ptrLast->SetDerivedFrom(ptrCopy);
      declType->SetBaseType( ptrCopy->GetBaseType() );
   }
   return sym;
}

/* ---------------------- class PtrType ----------------------------- */

#ifdef DEBUG
//##ModelId=4199F86A01D8
void FSPtrType::Print() const
{
   if (theQualifiers == FS_CONST) cout<<"const ";
   if (theQualifiers == FS_VOLATILE) cout<<"volatile ";
   cout<<"ptr to ";
   FSDeclarType::Print();
}
#endif

/* ------------------------ class ArrType ----------------------- */

#ifdef DEBUG
//##ModelId=4199F86A026C
void FSArrType::Print() const
{
   cout << "arr[" << ( dimPresent ? "dim" : "" ) << "] of ";
   FSDeclarType::Print();
}
#endif

/********************************************************************
 *             classes Symbol, SymbolList                           *
 ********************************************************************/

/* ------------------- class Symbol ---------------------------- */

char *FSUniqueName()
{
   static unsigned num = 0;
   static char name[] = "(unique)         ";
   itoa(++num,name+8,10);
   return name;
}

#ifdef DEBUG
//##ModelId=4199F86A03A2
void FSSymbol::Print() const
{
   cout<<" NAME:"<<aName<<" TYPE:";
   aType->Print();
   cout<<" DEFINED:"<<defined;
}
#endif

/* ----------------------- class SymbolList ---------------------- */

//##ModelId=4199F86B0124
Boolean FSSymbolList::LookUp(char *id, FSSymbol **ppSym) const
{
   Boolean found = FALSE;
   First();
   while ( More() )
   {
      if ( strcmp(id,Current()->GetId()) == 0 )
      {
         found = TRUE;
         *ppSym = Current();
         break;
      }
      Next();
   }
   return found;
}

#ifdef DEBUG
//##ModelId=4199F86B0136
void FSSymbolList::Print() const
{
   First();
   while ( More() )
   {
      Current()->Print();
      cout << ', ';
      Next();
   }
}
#endif
/* mutators */
//##ModelId=4199F86B0138
void FSSymbolList::Append( FSSymbol *sym)
{
   FSSymListElem *pSLElem = new FSSymListElem(sym);
   *last = pSLElem;
   last = &(pSLElem->next);
}
/* constructors & destructors */
//##ModelId=4199F86B014A
FSSymbolList::FSSymbolList(FSSymbolList &sList) :
first(0), last(&first), iter(0)
{ sList.First();
   while( sList.More() )
   {
      Append( new FSSymbol( *(sList.Current()) ) );
      sList.Next();
   }
}

//##ModelId=4199F86B0154
FSSymbolList::~FSSymbolList()
{
   FSSymListElem *pElem;
   First();
   while ( More() )
   {
      pElem = CurrElem();
      Next();
      delete pElem;
   }
}

/* -------------------- class FunctType -------------------------- */

#ifdef DEBUG
//##ModelId=4199F86B0212
void FSFunctType::Print() const
{
   cout<<" funct(";
   theArgs->Print();
   cout<<") returning ";
   FSDeclarType::Print();
}
#endif

/* ---------------- classes StrUnEnumType & StrUnEnumDefType -------------- */
#ifdef DEBUG
//##ModelId=4199F86B02A9
void FSStrUnEnumType::Print () const
{
   FSBaseType::Print();
   cout << defn->GetId();
}

//##ModelId=4199F86B0335
void FSStrUnEnumDefType::Print() const
{
   FSBaseType::Print();
   cout << "{ ";
   if( members ) members->Print();
   cout << "} ";
}
#endif

/**********************************************************************
 * class SymbolTable implementation                                   *
 **********************************************************************/

/* -------------------- class SymTabElem ---------------------------- */

#ifdef DEBUG
//##ModelId=4199F86C000E
void FSSymTabElem::Print() const
{
   cout<<"LEVEL:"<<GetSymbol()->GetLevel()<<' ';
   aSymbol->Print();
}
#endif

static const unsigned DEFTABLESIZE = 101;

//##ModelId=4199F86C0105
Boolean FSSymbolTable::LookUp(const char *id,FSSymbol **sym) const
{
   FSSymTabElem *ptrElem = hashPtrs[HashValue(id)];
   Boolean found;
   while ( ptrElem != 0 && (ptrElem->GetNameSpace() != FS_NS_ID ||
      strcmp(ptrElem->GetSymbol()->GetId(),id) != 0) )
      ptrElem = ptrElem->Next();
   found = (Boolean) (ptrElem != 0);
   if (found)
   {
      *sym = ptrElem->GetSymbol();
   }
#ifdef DEBUG
   cout << preProc->File() << '(' << preProc->Line() << ')';
   cout << "LOOKUP FOR "<<id;
   if ( found ) { ( *sym )->Print(); cout<<'\n'; }
   else cout << " NOT FOUND\n";
#endif
   return (found);
}

//##ModelId=4199F86C0111
Boolean FSSymbolTable::LookUpStrUnEnum(const char *id,FSSymbol **sym) const
{
   FSSymTabElem *ptrElem = hashPtrs[HashValue(id)];
   Boolean found;
   while ( ptrElem != 0 && (ptrElem->GetNameSpace() != FS_NS_STRUNENUM ||
      strcmp(ptrElem->GetSymbol()->GetId(),id) != 0) )
      ptrElem = ptrElem->Next();
   found = (Boolean) (ptrElem != 0);
   if (found)
   {
      *sym = ptrElem->GetSymbol();
   }
#ifdef DEBUG
   cout << preProc->File() << '(' << preProc->Line() << ')';
   cout << " TAG LOOKUP FOR "<<id;
   if ( found ) { ( *sym )->Print(); cout<<'\n'; }
   else cout << " NOT FOUND\n";
#endif
   return (found);
}

//##ModelId=4199F86C012E
void FSSymbolTable::Insert(FSSymbol *sym)
{
   /* treba dopunjavanje podrazumevanih specif. tipa
      i array u *
      extern !
   */
   FSSymTabElem **ptrChain = hashPtrs + HashValue(sym->GetId());
   FSSymTabElem *ptrElem = *ptrChain;
   /* look up to see if name already declared at the same level */
   while ( ptrElem != 0 && ptrElem->GetSymbol()->GetLevel() == Scope() &&
      (ptrElem->GetNameSpace() != FS_NS_ID ||
      strcmp(ptrElem->GetSymbol()->GetId(),sym->GetId()) != 0) )
      ptrElem = ptrElem->Next();
   if (ptrElem != 0 && ptrElem->GetSymbol()->GetLevel() == Scope() )
   {
      /* redeclaration at the same level */
      if ( Scope() != FS_EXT_DECL )
      {
         /* local symbols cannot be redeclared */
         FSAnzError( E_MULT_DECL_s, sym->GetId() );
         sym->SetDef( FS_DEFINED );
      }
      else if ( ptrElem->GetSymbol()->Defined() )
      {
         /* module level symbols can be redeclared, but not redefined*/
         /* !!! test za jednakost redeklaracije kod spoljnih simb. */
         FSAnzError( E_MULT_DEF_s, sym->GetId() );
      }
      delete ptrElem->GetSymbol();
      ptrElem->SetSymbol(sym);
   }
   else
   {
      sym->SetLevel(level);
      *ptrChain = new FSSymTabElem(sym,FS_NS_ID,*ptrChain);
      elemNum++; /* ovde eventualno prosirivanje vektora hashPtrs */
   }
#ifdef DEBUG
   cout << preProc->File() << '(' << preProc->Line() << ')';
   cout << " INSERT SYM ";
   sym->Print();
   cout << '\n';
#endif
}

//##ModelId=4199F86C0138
void FSSymbolTable::InsertStrUnEnum(FSSymbol **ppSym)
{
   /* poziva se uvek sa nekompletnim tipom; kompletiranje se vrsi izvan TS
      ako je redeklaracija vraca se stari simbol, a novi se brise
   */
   FSSymbol *sym = *ppSym;
   FSSymbol *oldSym;
   FSSymTabElem **ptrChain;
   if ( LookUpStrUnEnum(sym->GetId(), &oldSym) == FALSE ||
      oldSym->GetLevel() != Scope() )
   {
      ptrChain = hashPtrs + HashValue(sym->GetId());
      sym->SetLevel(level);
      *ptrChain = new FSSymTabElem(sym,FS_NS_STRUNENUM,*ptrChain);
      elemNum++; /* ovde eventualno prosirivanje vektora hashPtrs */
   }
   else
   {
      delete sym;
      *ppSym = oldSym;
   }
#ifdef DEBUG
   cout << preProc->File() << '(' << preProc->Line() << ')';
   cout << "INSERT TAG ";
   ( *ppSym )->Print();
   cout << '\n';
#endif
}

//##ModelId=4199F86C0141
void FSSymbolTable::EnterScope()
{
   level++;
#ifdef DEBUG
   cout << preProc->File() << '(' << preProc->Line() << ')';
   cout << "ENTERING SCOPE " << level << '\n';
#endif
}

//##ModelId=4199F86C0142
void FSSymbolTable::LeaveScope()
{
   unsigned hashPtrIndex;
   FSSymTabElem **ptrChain;
   FSSymTabElem *ptrElem;
   FSSymTabElem *ptrNextElem;
#ifdef DEBUG
   cout << preProc->File() << '(' << preProc->Line() << ')';
   cout<<"LEAVING SCOPE "<<level<<'\n';
   if ( level == FS_EXT_DECL ) Print();
#endif
   for(hashPtrIndex=0; hashPtrIndex < tableSize; hashPtrIndex++)
   {
      ptrChain = hashPtrs + hashPtrIndex;
      ptrElem = *ptrChain;

      // current scope symbols are at the beginning  of the list
      while( ptrElem != 0 && ptrElem->GetSymbol()->GetLevel() == level )
      {
         ptrNextElem = ptrElem->Next();
         delete ptrElem;
         ptrElem = ptrNextElem;
      }
      if (ptrElem != *ptrChain) *ptrChain = ptrElem;

      // now reset usedInCurrFunc flag for each global symbol if we
      // are leaving function body
      if ( level == FS_FUNC_BODY )
      {
         ptrElem = *ptrChain;

         while( ptrElem != 0 )
         {
            ptrElem->GetSymbol()->SetFirstUseInCurrFunc( TRUE );
            ptrElem = ptrElem->Next();
         }
      }
   }
   --level;
   return;
}

//##ModelId=4199F86C014B
FSSymbolTable::FSSymbolTable()
   : tableSize(DEFTABLESIZE),
     elemNum(0),
     level(FS_EXT_DECL)
{
   unsigned hashPtrsIndex;
   hashPtrs = new FSSymTabElem* [tableSize];
   for(hashPtrsIndex=0; hashPtrsIndex < tableSize; hashPtrsIndex++ )
      hashPtrs[hashPtrsIndex]=0;
}

//##ModelId=4199F86C014C
FSSymbolTable::~FSSymbolTable()
{
   unsigned hashPtrsIndex;
   FSSymTabElem *ptrElem;
   FSSymTabElem *ptrNextElem;
   for (hashPtrsIndex =0; hashPtrsIndex < tableSize; hashPtrsIndex++)
   {
      for(ptrElem=hashPtrs[hashPtrsIndex];ptrElem;ptrElem=ptrNextElem)
      {
         ptrNextElem = ptrElem->Next();
         delete ptrElem;
      }
   }
   delete [ /* tableSize */] hashPtrs;
}

//##ModelId=4199F86C0191
unsigned FSSymbolTable::HashValue(const char *id) const
{
   unsigned h;
   for (h=0; *id; id++) h += *id;
   return (h % tableSize);
}

#ifdef DEBUG
//##ModelId=4199F86C0124
void FSSymbolTable::Print() const
{
   unsigned i;
   FSSymTabElem *ptrElem;
   cout<<"**************SYMBOL TABLE***********************\n";
   for (i=0; i < tableSize ; i++ )
   {
      if ( hashPtrs[i] ) cout<<"hash="<<i<<'\n';
      for(ptrElem=hashPtrs[i]; ptrElem ; ptrElem = ptrElem->Next() )
      {
         cout << "           ";
         ptrElem->Print();
         cout << '\n';
      }
   }
}
#endif

/* --------------------- Miscellany --------------------------- */


FSSymbolTable *fsSymTable;

void FSInsertSymListtoTable(FSSymbolTable &SymTab, FSSymbolList &SymList)
{
   FSSymbol *pSym;
   SymList.First();
   while ( SymList.More() )
   {
      pSym = new FSSymbol( *(SymList.Current()) );
      SymTab.Insert( pSym );
      SymList.Next();
   }
}

void FSPatchArgs( FSSymbol *fct, FSSymbol *arg )
{
   /* za K & R def. fja, proverava slaganje argumenata i popravlja fct listu
      u skladu sa datim argumentom
   */
   FSFunctType *fctType = (FSFunctType *) ( fct->GetType() );
   FSSymbolList *fctArgs = fctType->GetArgs();
   FSSymbol *fctArg;
   if ( fctType->GetStyle() == FS_S_ANSI )
   {
      FSAnzWarning( W_OLD_ANSI_FUNC_DECS_s, fct->GetId() );
   }
   else /* K & R style function declaration */
      if ( fctArgs->LookUp( arg->GetId(), &fctArg ) == TRUE )
      {
         if (fctArg->Defined()==FS_DEFINED) FSAnzError( E_MULT_DECL_s, fctArg->GetId() );
         fctArg->SetDef(FS_DEFINED);
         delete fctArg->GetType();
         fctArg->SetType(arg->GetType()->Copy());
      }
      else
         FSAnzError( E_NOT_FUNC_PARAM_s, arg->GetId() );
   delete arg;
}


/*
int main() {
  Type *t1 = new BaseType(INT);
  Type *t2 = new BaseType(FLOAT);
  Symbol *s1 = new Symbol("Symbol 1",t1);
  Symbol *s2 = new Symbol("Symbol 2",t2);
  SymbolList sl;
  sl.Append(s1);
  sl.Append(s2);
  sl.Print();
  return 0;
}

int main () {
  SymbolTable *ST = new SymbolTable;
  unsigned i,j;
  char line[80];
  for (i=0; i < 3 ; i++ ) {
    ST->EnterScope();
    for (j=0; j < 3; j++ ) {
      cin>>line;
      Type *tp1= new BaseType(CONST);
      ((BaseType *)(tp1->GetBaseType()))->AddSpecifier(INT);
      Symbol *sym1 = new Symbol(line,tp1);
      ST->Insert(sym1);
    }
  }
  ST->Print();
  ST->LeaveScope();
  ST->LeaveScope();
  ST->Print();
  delete ST;
  return 0;
}
*/

