//----------------------------------------------------------------------------
//  Description :
//      type representation and symbol table classes interface
//----------------------------------------------------------------------------
//  Changes:
//  $Log__$
//----------------------------------------------------------------------------

#ifndef FSSYMTAB_H
#define FSSYMTAB_H

#if !defined ( _INC_STRING ) || !defined( FSBOOL_H )
   #error string.h and fsbool.h required!
#endif

/*********************************************************************
 *     Klase vezane za predstavljanje tipova                         *
 *********************************************************************/

#pragma warning( disable : 4355 )
//##ModelId=4199F8690219
enum FSTypeId {
	FS_T_BASE, 
	FS_T_ENUM, 
	FS_T_STRUCT, 
	FS_T_UNION, 
	FS_T_PTR, 
	FS_T_FUNCT, 
	FS_T_ARR
};

/* --------------------abstract class Type ------------------------ */

//##ModelId=4199F8690269
class FSType
{
public:
   /* selectors */
	//##ModelId=4199F8690273
   const FSTypeId GetId() const { return anId; }
	//##ModelId=4199F869027D
   FSType *GetBaseType() const {return theBase;}
	//##ModelId=4199F869027F
   virtual FSType *Copy() const = 0;
#ifdef DEBUG
	//##ModelId=4199F8690288
   virtual void Print() const = 0;
#endif
   /* modifiers */
	//##ModelId=4199F8690291
   void SetBaseType( FSType *pBase ) { theBase = pBase; }
   /* constructors & destructors */
	//##ModelId=4199F869029C
   FSType( FSTypeId typId, FSType *pBase ) : anId( typId ), theBase( pBase ) {}
	//##ModelId=4199F86902A6
   virtual ~FSType() {}
private:
	//##ModelId=4199F86902A9
   const FSTypeId anId;
	//##ModelId=4199F86902BD
   FSType *theBase; /* vazeci samo u tipu na najvisem nivou slozenog tipa */
};

/* ----------------------- class BaseType ------------------------- */

//##ModelId=4199F869033C
enum FSBaseTypeSpecifiers
{
   /* type specifiers */
	//##ModelId=4199F8690347
   FS_NOTYPE = 0,
	//##ModelId=4199F8690350
   FS_INT = 1,
	//##ModelId=4199F869035A
   FS_CHAR = 2,
	//##ModelId=4199F8690364
   FS_FLOAT = 4,
	//##ModelId=4199F869036E
   FS_DOUBLE = 8,
	//##ModelId=4199F8690378
   FS_VOID = 16,
	//##ModelId=4199F8690382
   FS_LONG = 32,
	//##ModelId=4199F869038C
   FS_SHORT = 64,
	//##ModelId=4199F8690396
   FS_SIGNED = 128,
	//##ModelId=4199F86903A0
   FS_UNSIGNED = 256,
   /* type qualifiers */
	//##ModelId=4199F86903AA
   FS_CONST = 512,
	//##ModelId=4199F86903B4
   FS_VOLATILE = 1024,
   /* class specifiers */
	//##ModelId=4199F86903BE
   FS_AUTO = -1,
	//##ModelId=4199F86903C8
   FS_EXTERN = -2,
	//##ModelId=4199F86903D2
   FS_REGISTER = -4,
	//##ModelId=4199F86903DC
   FS_STATIC = -8,
	//##ModelId=4199F86903E6
   FS_TYPEDEF = -16
};

//##ModelId=4199F86A0076
class FSBaseType : public FSType
{
public:
   /* selectors */
#ifdef DEBUG
	//##ModelId=4199F86A0081
   void Print() const;
#endif
	//##ModelId=4199F86A008B
   FSType *Copy() const { return new FSBaseType( *this ); }
	//##ModelId=4199F86A0094
   Boolean GetSpecifier( FSBaseTypeSpecifiers spec )
   {
      return ( spec > 0 ) ? ( Boolean ) ( theType & spec )
      : ( Boolean ) ( theClass & -spec );
   }
	//##ModelId=4199F86A0096
   void CheckTypeDef( Boolean &tDefAllowed ) const;
   /* mutators */
	//##ModelId=4199F86A00A0
   static FSType *Merge( FSType *dst,FSType *src );
	//##ModelId=4199F86A00B2
   void AddSpecifier( FSBaseTypeSpecifiers spec );
	//##ModelId=4199F86A00BC
   void RemoveSpecifier( FSBaseTypeSpecifiers spec );

   /* constructors & destructors */
	//##ModelId=4199F86A00C6
   FSBaseType( FSTypeId typId = FS_T_BASE ) : FSType( typId,this ), theType( 0 ),
   theClass( 0 ) {}
	//##ModelId=4199F86A00D0
   FSBaseType( FSBaseTypeSpecifiers spec, FSTypeId typId = FS_T_BASE ) : FSType( typId,
   this ),
   theType( 0 ), theClass( 0 )
   { AddSpecifier( spec );
   }
	//##ModelId=4199F86A00DA
   FSBaseType( const FSBaseType &bt ) : FSType( FS_T_BASE,this ),
   theType( bt.theType ), theClass( bt.theClass ) {}
private:
	//##ModelId=4199F86A00E5
   unsigned theType;
	//##ModelId=4199F86A00EF
   unsigned theClass;
};

/* ------------ abstract class DeclarType ------------------------ */

//##ModelId=4199F86A0134
class FSDeclarType : public FSType
{
public:
   /* selectors */
	//##ModelId=4199F86A0148
   virtual FSType *Copy() const = 0;
	//##ModelId=4199F86A014A
   FSType *DerivedFrom() const { return derivedFrom; }
#ifdef DEBUG
	//##ModelId=4199F86A0153
   virtual void Print() const { derivedFrom->Print(); }
#endif
   /* modifiers */
	//##ModelId=4199F86A015C
   void SetDerivedFrom( FSType *ptrDer ) { derivedFrom = ptrDer; }
   /* constructors & destructors */
	//##ModelId=4199F86A0167
   FSDeclarType( FSTypeId typId, FSType *pBase ) :
   FSType( typId,pBase ), derivedFrom( 0 ) {}
	//##ModelId=4199F86A0171
   FSDeclarType( const FSDeclarType &dt ) :
   FSType( dt.GetId(),0 ),
   derivedFrom( ( dt.derivedFrom == 0 ) ? 0 : dt.derivedFrom->Copy() )
   { FSType::SetBaseType( ( derivedFrom == 0 ) ? this : derivedFrom->GetBaseType
      () ) ;
   }
	//##ModelId=4199F86A017B
   virtual ~FSDeclarType() { delete derivedFrom; }
private:
	//##ModelId=4199F86A017E
   FSType *derivedFrom;
};

/* ---------------------- class PtrType ------------------------ */

//##ModelId=4199F86A01C1
class FSPtrType : public FSDeclarType
{
public:
   /* selectors */
	//##ModelId=4199F86A01CC
   FSType *Copy() const { return new FSPtrType( *this ); }
	//##ModelId=4199F86A01D6
   FSBaseTypeSpecifiers GetQualifiers() const { return theQualifiers; }
#ifdef DEBUG
	//##ModelId=4199F86A01D8
   void Print() const;
#endif
   /* constructors & destructors */
	//##ModelId=4199F86A01E0
   FSPtrType( FSBaseTypeSpecifiers sp = FS_NOTYPE ) :
   FSDeclarType( FS_T_PTR, this ), theQualifiers( sp ) {}
	//##ModelId=4199F86A01EA
   FSPtrType( const FSPtrType &pTp ) : FSDeclarType( pTp ),
   theQualifiers( pTp.theQualifiers ) {}

private:
	//##ModelId=4199F86A01F4
   FSBaseTypeSpecifiers theQualifiers;
};

/* ------------------------ class ArrType ----------------------- */

//##ModelId=4199F86A0207
enum FSDimPresentFlag {
	FS_NO_DIM_PRESENT,
	FS_DIM_PRESENT
};

//##ModelId=4199F86A024D
class FSArrType : public FSDeclarType
{
public:
   /* selectors */
	//##ModelId=4199F86A0261
   FSType *Copy() const { return new FSArrType( *this ); }
	//##ModelId=4199F86A0263
   FSDimPresentFlag DimPresent() const { return dimPresent; }
#ifdef DEBUG
	//##ModelId=4199F86A026C
   void Print() const ;
#endif
   /* constructors & destructors */
	//##ModelId=4199F86A0275
   FSArrType( FSDimPresentFlag dim = FS_NO_DIM_PRESENT ) :
   FSDeclarType( FS_T_ARR,this ), dimPresent( dim ) {}
	//##ModelId=4199F86A0277
   FSArrType( const FSArrType &aTp ) : FSDeclarType( aTp ),
   dimPresent( aTp.dimPresent ) {}

private:
	//##ModelId=4199F86A0281
   FSDimPresentFlag dimPresent;
};

/********************************************************************
 *             classes Symbol, SymbolList                           *
 ********************************************************************/

/* ------------------- class Symbol ---------------------------- */

//##ModelId=4199F86A029D
typedef int FSNestingLevel;
const FSNestingLevel FS_NOT_IN_TABLE = 0;
const FSNestingLevel FS_EXT_DECL = 1;
const FSNestingLevel FS_FUNC_BODY = 2;
//##ModelId=4199F86A02B1
enum FSDefinedFlag {
	FS_NOT_DEFINED,
	FS_DEFINED
};
//##ModelId=4199F86A02CF
enum FSCopyFlag {
	FS_NO_ID_COPY,
	FS_ID_COPY
};

//##ModelId=4199F86A035B
class FSSymbol
{
   public:

      /* constructors & destructor */
	//##ModelId=4199F86A0365
      FSSymbol ( char *id, FSCopyFlag cp = FS_ID_COPY, FSType *ty = 0,
                 FSDefinedFlag def = FS_NOT_DEFINED
               )
         : aName( ( cp == FS_ID_COPY ) ? strcpy( new char[strlen( id )+1],id ) : id ),
           aType( ty ),
           defined( def ),
           aLevel( FS_NOT_IN_TABLE ),
           firstUseInCurrFunc( TRUE )
      {}

	//##ModelId=4199F86A0379
      FSSymbol( const FSSymbol &orig )
         : aName( new char[strlen( orig.aName )+1] ),
           aType( orig.aType->Copy() ),
           defined( orig.defined ),
           aLevel( orig.aLevel ),
           firstUseInCurrFunc( orig.firstUseInCurrFunc )
      {
         strcpy( aName,orig.aName );
      }

	//##ModelId=4199F86A037B
      ~FSSymbol () { delete [ /* strlen( aName )+1 */] aName; delete aType; }

      /* selectors */

	//##ModelId=4199F86A0383
      char *GetId() const {return aName;}

	//##ModelId=4199F86A038D
      FSType *GetType() const {return aType;}

	//##ModelId=4199F86A038F
      FSDefinedFlag Defined() const {return defined;}

	//##ModelId=4199F86A0398
      FSNestingLevel GetLevel() const {return aLevel;}

	//##ModelId=4199F86A039A
      Boolean FirstUseInCurrFunc() const { return firstUseInCurrFunc; }

#ifdef DEBUG
	//##ModelId=4199F86A03A2
      void Print() const ;
#endif

      /* modifiers */

	//##ModelId=4199F86A03AB
      void SetType( FSType *tp ) {aType = tp; }

	//##ModelId=4199F86A03B5
      void SetDef( FSDefinedFlag df ) {defined = df;}

	//##ModelId=4199F86A03BF
      void SetLevel( FSNestingLevel lv ) { aLevel = lv; }

	//##ModelId=4199F86A03C1
      void SetFirstUseInCurrFunc( Boolean used ) { firstUseInCurrFunc = used; }

   private:

	//##ModelId=4199F86A03CA
      char *aName;
	//##ModelId=4199F86A03DE
      FSType *aType;
	//##ModelId=4199F86A03E8
      FSDefinedFlag defined;
	//##ModelId=4199F86B0014
      FSNestingLevel aLevel;
	//##ModelId=4199F86B001E
      Boolean firstUseInCurrFunc; // TRUE for first use, FALSE for next uses
};

char *FSUniqueName() ;

/* ----------------------- class SymbolList ---------------------- */

//##ModelId=4199F86B0050
struct FSSymListElem
{
	//##ModelId=4199F86B0065
   FSSymbol *ptrSym;
	//##ModelId=4199F86B0071
   FSSymListElem *next;
   /* constructors & destructors */
	//##ModelId=4199F86B0078
   FSSymListElem( FSSymbol *pSym, FSSymListElem *nx = 0 ) :
   ptrSym( pSym ), next( nx ) {}
	//##ModelId=4199F86B008C
   ~FSSymListElem() { delete ptrSym; }
};

//##ModelId=4199F86B00FA
class FSSymbolList
{
public:
   /* selectors */
	//##ModelId=4199F86B010E
   void Next() const
   {
      ( FSSymListElem * )iter = iter->next;
   }
	//##ModelId=4199F86B0110
   Boolean More() const
   {
      return ( Boolean ) ( iter != 0 );
   }
	//##ModelId=4199F86B0119
   FSSymbol *Current() const { return iter->ptrSym; }
	//##ModelId=4199F86B0122
   void First() const { ( FSSymListElem * )iter = first; }
	//##ModelId=4199F86B0124
   Boolean LookUp( char *id, FSSymbol **ppSym ) const ;

#ifdef DEBUG
	//##ModelId=4199F86B0136
   void Print() const ;
#endif
   /* mutators */
	//##ModelId=4199F86B0138
   void Append( FSSymbol *sym ) ;
   /* constructors & destructors */
	//##ModelId=4199F86B0141
   FSSymbolList() : first( 0 ), last( &first ), iter( 0 ) {}
	//##ModelId=4199F86B014A
   FSSymbolList( FSSymbolList &sList );
	//##ModelId=4199F86B0154
   ~FSSymbolList() ;
private:
	//##ModelId=4199F86B0156
   FSSymListElem *first;
	//##ModelId=4199F86B0169
   FSSymListElem **last;
	//##ModelId=4199F86B0173
   FSSymListElem *iter;
   /* selectors */
	//##ModelId=4199F86B0186
   FSSymListElem *CurrElem() const {return iter;}
};

/* -------------------- class FunctType -------------------------- */

//##ModelId=4199F86B019A
enum FSFunctArgsStyle {
	FS_S_KR, 
	FS_S_ANSI
};

//##ModelId=4199F86B01F4
class FSFunctType : public FSDeclarType
{
public:
   /* selectors */
	//##ModelId=4199F86B0208
   FSType *Copy() const { return new FSFunctType( *this ); }
#ifdef DEBUG
	//##ModelId=4199F86B0212
   void Print() const ;
#endif
	//##ModelId=4199F86B0214
   FSFunctArgsStyle GetStyle() const { return theStyle; }
	//##ModelId=4199F86B021D
   FSSymbolList *GetArgs() const { return theArgs; }
   /* constructors & destructors */
	//##ModelId=4199F86B021F
   FSFunctType( enum FSFunctArgsStyle st = FS_S_ANSI, FSSymbolList *sLe = new
   FSSymbolList ) :
   FSDeclarType( FS_T_FUNCT, this ), theStyle( st ), theArgs( sLe ) {}
	//##ModelId=4199F86B0230
   FSFunctType( const FSFunctType &pFt ) : FSDeclarType( pFt ), theStyle( pFt.
   theStyle ),
   theArgs( new FSSymbolList( *( pFt.theArgs ) ) ) {}
	//##ModelId=4199F86B0232
   ~FSFunctType() { delete theArgs; }
private:
	//##ModelId=4199F86B023B
   FSFunctArgsStyle theStyle;
	//##ModelId=4199F86B024E
   FSSymbolList *theArgs;
};

/* ---------------- classes StrUnEnumType & StrUnEnumDefType -------------- */

//##ModelId=4199F86B028A
class FSStrUnEnumType : public FSBaseType
{
public:
   /* selectors */
	//##ModelId=4199F86B029E
   FSType *Copy() const { return new FSStrUnEnumType( *this ); }
	//##ModelId=4199F86B02A0
   FSSymbol *GetDefinition() const { return defn; }
#ifdef DEBUG
	//##ModelId=4199F86B02A9
   void Print () const ;
#endif
   /* constructors & destructors */
	//##ModelId=4199F86B02B2
   FSStrUnEnumType( FSTypeId typId, FSSymbol *def, FSBaseTypeSpecifiers spec = FS_NOTYPE
   ) :
   FSBaseType( spec,typId ), defn( def ) {}
	//##ModelId=4199F86B02BD
   FSStrUnEnumType( const FSStrUnEnumType &orig ) :
   FSBaseType( orig.GetId() ), defn( orig.defn ) {}
private:
	//##ModelId=4199F86B02C7
   FSSymbol *defn;
};

//##ModelId=4199F86B0317
class FSStrUnEnumDefType : public FSBaseType
{
public:
   /* selectors */
	//##ModelId=4199F86B0322
   FSSymbolList *GetMembers() const { return members; }
	//##ModelId=4199F86B032C
   FSType *Copy() const { return new FSStrUnEnumDefType( *this ); }
#ifdef DEBUG
	//##ModelId=4199F86B0335
   void Print() const ;
#endif
   /* mutators */
	//##ModelId=4199F86B0337
   void SetMembers( FSSymbolList *newMem ) { members = newMem; }
   /* constructors & destructors */
	//##ModelId=4199F86B0340
   FSStrUnEnumDefType( FSTypeId typId, FSSymbolList *mem = 0 ) :
   FSBaseType( typId ), members( mem ) {}
	//##ModelId=4199F86B034A
   FSStrUnEnumDefType( const FSStrUnEnumDefType &orig ) :
   FSBaseType( orig.GetId() ), members( orig.members ) {}
	//##ModelId=4199F86B0354
   ~FSStrUnEnumDefType() { delete members; }
private:
	//##ModelId=4199F86B035E
   FSSymbolList *members;
};


/**********************************************************************
 *      Klase za tabelu simbola                                       *
 **********************************************************************/

//##ModelId=4199F86B0371
enum FSNameSpaceType {
	FS_NS_ID, 
	FS_NS_STRUNENUM
};

//##ModelId=4199F86B03D5
class FSSymTabElem
{
public:
   /* selectors */
	//##ModelId=4199F86B03DF
   FSSymbol *GetSymbol() const { return ( FSSymbol * ) aSymbol;}
	//##ModelId=4199F86C0002
   FSSymTabElem *Next() const {return nextPtr;}
	//##ModelId=4199F86C000C
   FSNameSpaceType GetNameSpace() const { return nameSpace; }
#ifdef DEBUG
	//##ModelId=4199F86C000E
   void Print() const ;
#endif
   /* mutators */
	//##ModelId=4199F86C0016
   void SetNext( FSSymTabElem *ptr ) {nextPtr = ptr;}
	//##ModelId=4199F86C0020
   void SetSymbol( FSSymbol *sym ) { aSymbol = sym; }
   /* constructors & destructors */
	//##ModelId=4199F86C002A
   FSSymTabElem( const FSSymbol *sy, FSNameSpaceType nsp, FSSymTabElem *ptr ) :
   aSymbol( sy ), nameSpace( nsp ), nextPtr( ptr ) {}
	//##ModelId=4199F86C003D
   ~FSSymTabElem() { delete aSymbol;}
private:
	//##ModelId=4199F86C003F
   const FSSymbol *aSymbol;
	//##ModelId=4199F86C0052
   FSNameSpaceType nameSpace;
	//##ModelId=4199F86C0065
   FSSymTabElem *nextPtr;
};

//##ModelId=4199F86C00F1
class FSSymbolTable
{
public:
   /* selectors */
	//##ModelId=4199F86C0105
   Boolean LookUp( const char *id, FSSymbol **sym ) const;
	//##ModelId=4199F86C0111
   Boolean LookUpStrUnEnum( const char *id, FSSymbol **sym ) const;
	//##ModelId=4199F86C0123
   FSNestingLevel Scope() {return level;}
#ifdef DEBUG
	//##ModelId=4199F86C0124
   void Print() const;
#endif
   /* mutators */
	//##ModelId=4199F86C012E
   void Insert( FSSymbol *sym );
	//##ModelId=4199F86C0138
   void InsertStrUnEnum( FSSymbol **sym );

	//##ModelId=4199F86C0141
   void EnterScope();

	//##ModelId=4199F86C0142
   void LeaveScope();
   /* constructors & destructors */
	//##ModelId=4199F86C014B
   FSSymbolTable();
	//##ModelId=4199F86C014C
   ~FSSymbolTable();
private:
	//##ModelId=4199F86C0156
   unsigned tableSize;
	//##ModelId=4199F86C0160
   FSSymTabElem **hashPtrs;
	//##ModelId=4199F86C0174
   unsigned elemNum;
	//##ModelId=4199F86C017E
   FSNestingLevel level;
	//##ModelId=4199F86C0191
   unsigned HashValue( const char *id ) const;

};

/* --------------------------- Miscellany ------------------------ */

extern FSSymbolTable *fsSymTable;

void FSInsertSymListtoTable( FSSymbolTable &SymTab, FSSymbolList &SymList );

//##ModelId=4199F86C01AF
enum FSCopySpecFlag {
	FS_NO_SPEC_COPY,
	FS_SPEC_COPY
};

FSSymbol *FSAppendType( FSType *specType, FSSymbol *sym,FSCopySpecFlag cpySpec =
FS_NO_SPEC_COPY );

void FSPatchArgs( FSSymbol *fct, FSSymbol *arg );

#endif

